'''The :mod:`pulsar.apps.pulse` module is a :django:`django application <>`
for running a django web site with pulsar.
Add it to the list of your ``INSTALLED_APPS``::

    INSTALLED_APPS = (
        ...,
        'pulsar.apps.pulse',
        ...
    )

and run the site via the ``pulse`` command::

    python manage.py pulse

Check the :ref:`django chat example <tutorials-django>` for a django chat
application served by a multiprocessing pulsar server.

By default, the ``pulse`` command creates a :class:`Wsgi` middleware which
runs the django application in a separate thread of execution from the
main event loop.
This is a standard programming pattern when using :ref:`asyncio with blocking
functions <asyncio-multithreading>`.
To control the number of thread workers in the event loop executor (which
is a pool of threads) one uses the
:ref:`thread-workers <setting-thread_workers>` option. For example, the
following command::

    python manage.py pulse -w 4 --thread-workers 20

will run four :ref:`process based actors <concurrency>`, each with
an executor with up to 20 threads.

Greenlets
===============

It is possible to run django in fully asynchronous mode, i.e. without
running the middleware in the event loop executor.
Currently, this is available when using :postgresql:`PostgreSql backend <>`
only, and it requires the :greenlet:`greenlet library <>`.

To run django using greenlet support::

    python manage.py pulse -w 4 --greenlet

By default it will run the django middleware on a pool of 100 greenlets (and
therefore approximately 100 separate database connections per actor). To
adjust this number::

    python manage.py pulse -w 4 --greenlet 200


Wsgi middleware
===================

.. autoclass:: Wsgi
   :members:
   :member-order: bysource
'''
from pulsar.apps.wsgi import (LazyWsgi, WsgiHandler,
                              wait_for_body_middleware,
                              middleware_in_executor)
try:
    from pulsar.apps import greenio
    from pulsar.apps.greenio import pg, local
except ImportError:
    greenio = None
    pg = None


class Wsgi(LazyWsgi):
    '''The Wsgi middleware used by the django ``pulse`` command
    '''
    cfg = None

    def setup(self, environ=None):
        '''Set up the :class:`.WsgiHandler` the first time this
        middleware is accessed.
        '''
        from django.conf import settings
        from django.core.wsgi import get_wsgi_application
        #
        app = get_wsgi_application()
        green_workers = self.cfg.greenlet if self.cfg else 0
        if greenio and green_workers:
            if pg:
                pg.make_asynchronous()
            app = greenio.RunInPool(app, green_workers)
            self.green_safe_connections()
        else:
            app = middleware_in_executor(app)
        return WsgiHandler((wait_for_body_middleware, app))

    def green_safe_connections(self):
        from django.db import connections
        connections._connections = local()



'''
This section describes the asynchronous WSGI specification used by pulsar.
It is a superset of the `WSGI 1.0.1`_ specification for synchronous
server/middleware.
If an application handler is synchronous, this specification is exactly
equivalent to `WSGI 1.0.1`_. Changes with respect `WSGI 1.0.1`_
concern asynchronous responses and nothing else.

Introduction
========================

The WSGI interface has two sides: the ``server`` or ``gateway`` side, and the
``application`` or ``framework`` side. The server side invokes a callable
object, here referred as **application handler**, that is provided by the
application side.


.. note::

    A standard WSGI application handler is always a callable, either a function
    or a callable object, which accepts two positional arguments:
    ``environ`` and ``start_response``. When called by the server,
    the application object must return an iterable yielding zero or more bytes.


.. _wsgi-handlers:

Application handlers
=============================

An asynchronous :ref:`application handler <wsgi-handlers>` must conform
with the standard `WSGI 1.0.1`_ specification with the following two
exceptions:

* It can return a :class:`~asyncio.Future`.
* If it returns a :class:`~asyncio.Future`, it must results in an
  :ref:`asynchronous iterable <wsgi-async-iter>`.

Pulsar is shipped with two WSGI application handlers documented below.

.. _wsgi-async-iter:

Asynchronous Iterable
========================

An asynchronous iterable is an iterable over a combination of ``bytes`` or
:class:`~asyncio.Future` which result in ``bytes``.
For example this could be an asynchronous iterable::

    def simple_async():
        yield b'hello'
        c = Future()
        c.set_result(b' ')
        yield c
        yield b'World!'


.. _wsgi-lazy-handler:

WsgiHandler
======================

The first application handler is the :class:`WsgiHandler`
which is a step above the :ref:`hello callable <tutorials-hello-world>`
in the tutorial. It accepts two iterables, a list of
:ref:`wsgi middleware <wsgi-middleware>` and an optional list of
:ref:`response middleware <wsgi-response-middleware>`.

.. autoclass:: WsgiHandler
   :members:
   :member-order: bysource


.. _wsgi-handler:

Lazy Wsgi Handler
======================

.. autoclass:: LazyWsgi
   :members:
   :member-order: bysource


.. _wsgi-pulsar-variables:

Pulsar Variables
======================
Pulsar inject two server-defined variables into the WSGI environ:

* ``pulsar.connection``, the :class:`.Connection` serving the request
* ``pulsar.cfg``, the :class:`.Config` dictionary of the server

The event loop serving the application can be retrieved from the connection
via the ``_loop`` attribute::

    loop = environ['pulsar.connection']._loop

.. _WSGI: http://www.wsgi.org
.. _`WSGI 1.0.1`: http://www.python.org/dev/peps/pep-3333/
'''
import sys
import types

from pulsar import async, Http404, coroutine_return, appengine
from pulsar.utils.structures import OrderedDict
from pulsar.utils.log import LocalMixin, local_method

from .utils import handle_wsgi_error, wsgi_request
from .wrappers import WsgiResponse


__all__ = ['WsgiHandler', 'LazyWsgi']


class WsgiHandler(object):
    '''An handler for application conforming to python WSGI_.

    .. attribute:: middleware

        List of :ref:`asynchronous WSGI middleware <wsgi-middleware>` callables
        which accept ``environ`` and ``start_response`` as arguments.
        The order matter, since the response returned by the callable
        is the non ``None`` value returned by a middleware.

    .. attribute:: response_middleware

        List of functions of the form::

            def ..(environ, response):
                ...

        where ``response`` is a :ref:`WsgiResponse <wsgi-response>`.
        Pulsar contains some
        :ref:`response middlewares <wsgi-response-middleware>`.

    '''
    def __init__(self, middleware=None, response_middleware=None, **kwargs):
        if middleware:
            middleware = list(middleware)
        self.middleware = middleware or []
        self.response_middleware = response_middleware or []
        self.error_handlers = {}

    def __call__(self, environ, start_response):
        '''The WSGI callable'''
        environ['error.handlers'] = self.error_handlers
        if appengine:
            return self._appengine(environ, start_response)
        else:
            c = environ.get('pulsar.connection')
            loop = c._loop if c else None
            return async(self._call(environ, start_response), loop)

    def _call(self, environ, start_response):
        resp = None
        try:
            for middleware in self.middleware:
                resp = yield middleware(environ, start_response)
                if resp is not None:
                    break
            if resp is None:
                raise Http404
        except Exception as exc:
            resp = yield handle_wsgi_error(environ, exc)
        if isinstance(resp, WsgiResponse):
            # The response is a WSGIResponse
            for middleware in self.response_middleware:
                resp = yield middleware(environ, resp)
            start_response(resp.status, resp.get_headers())
        coroutine_return(resp)

    def _appengine(self, environ, start_response):
        resp = None
        try:
            for middleware in self.middleware:
                resp = middleware(environ, start_response)
                if resp is not None:
                    break
            if resp is None:
                raise Http404
        except Exception as exc:
            resp = handle_wsgi_error(environ, exc)
        if isinstance(resp, WsgiResponse):
            # The response is a WSGIResponse
            for middleware in self.response_middleware:
                resp = middleware(environ, resp)
            start_response(resp.status, resp.get_headers())
        return resp


class LazyWsgi(LocalMixin):
    '''A :ref:`wsgi handler <wsgi-handlers>` which loads the actual
    handler the first time it is called.

    Subclasses must implement the :meth:`setup` method.
    Useful when working in multiprocessing mode when the application
    handler must be a ``picklable`` instance. This handler can rebuild
    its wsgi :attr:`handler` every time is pickled and un-pickled without
    causing serialisation issues.
    '''
    def __call__(self, environ, start_response):
        return self.handler(environ)(environ, start_response)

    @local_method
    def handler(self, environ=None):
        '''The :ref:`wsgi application handler <wsgi-handlers>` which
        is loaded via the :meth:`setup` method, once only,
        when first accessed.
        '''
        return self.setup(environ)

    def setup(self, environ=None):
        '''The setup function for this :class:`LazyWsgi`.

        Called once only the first time this application handler is invoked.
        This **must** be implemented by subclasses and **must** return a
        :ref:`wsgi application handler <wsgi-handlers>`.
        '''
        raise NotImplementedError



'''
This example illustrates how to write a UDP Echo server and client pair.
The code for this example is located in the :mod:`examples.echoudp.manage`
module.

Run The example
====================

To run the server::

    python manage.py

Open a new shell, in this directory, launch python and type::

    >>> from manage import Echo
    >>> echo = Echo(('localhost',8060))
    >>> echo(b'Hello!')
    b'Hello!'

Writing the Client
=========================

The first step is to write a small class handling a connection
pool with the remote server. The :class:`Echo` class does just that,
it subclass the handy :class:`.AbstractUdpClient` and uses
the asynchronous :class:`.Pool` of connections as backbone.

The second step is the implementation of the :class:`EchoUdpProtocol`,
a subclass of :class:`.DatagramProtocol`.
The :class:`EchoUdpProtocol` is needed for two reasons:

* It encodes and sends the request to the remote server
* It listens for incoming data from the remote server via the
  :meth:`~EchoUdpProtocol.datagram_received` method.



Implementation
==================

Echo Udp Protocol
~~~~~~~~~~~~~~~~~~~~~~~~~~~

.. autoclass:: EchoUdpProtocol
   :members:
   :member-order: bysource


Echo Client
~~~~~~~~~~~~~~~~~~~~~~~~~~~

.. autoclass:: Echo
   :members:
   :member-order: bysource

   .. automethod:: __call__

Echo Server
~~~~~~~~~~~~~~~~~~~~~~~~~~~

.. autofunction:: server

'''
try:
    import pulsar
except ImportError:     # pragma nocover
    import sys
    sys.path.append('../../')
    import pulsar

from pulsar import coroutine_return, Pool, Future, DatagramProtocol, task
from pulsar.utils.pep import to_bytes
from pulsar.apps.socket import UdpSocketServer


class EchoUdpProtocol(DatagramProtocol):
    '''A base :class:`.DatagramProtocol` for UDP echo clients and servers.

    The only difference between client and server is the implementation
    of the :meth:`response` method.
    '''
    separator = b'\r\n\r\n'
    '''A separator for messages.'''
    buffer = None
    '''The buffer for long messages'''

    def popbuffer(self, addr):
        if self.buffer:
            return self.buffer.pop(addr, None)

    def datagram_received(self, data, addr):
        '''Handle data from ``addr``.
        '''
        while data:
            idx = data.find(self.separator)
            if idx >= 0:    # we have a full message
                idx += len(self.separator)
                chunk, data = data[:idx], data[idx:]
                buffer = self.popbuffer(addr)
                self.response(buffer + chunk if buffer else chunk, addr)
            else:
                if self.buffer is None:
                    self.buffer = {}
                if addr in self.buffer:
                    self.buffer[addr] += data
                else:
                    self.buffer[addr] = data

    def response(self, data, addr):
        '''Abstract response handler'''
        raise NotImplementedError


class EchoUdpClientProtocol(EchoUdpProtocol):
    _waiting = None

    def send(self, message):
        assert self._waiting is None
        self._waiting = d = Future(loop=self._loop)
        self._transport.sendto(to_bytes(message)+self.separator)
        return d

    def response(self, data, addr):
        '''Got a full response'''
        d, self._waiting = self._waiting, None
        if d:
            d.set_result(data[:-len(self.separator)])


class EchoUdpServerProtocol(EchoUdpProtocol):
    '''The :class:`EchoUdpProtocol` used by the echo udp :func:`server`.
    '''
    def response(self, data, addr):
        '''Override :meth:`~EchoProtocol.response` method by writing the
        ``data`` received back to the client.
        '''
        self._transport.sendto(data, addr)


class Echo(pulsar.AbstractUdpClient):
    '''A client for the echo server.

    :param address: set the :attr:`address` attribute
    :param pool_size: used when initialising the connetion :attr:`pool`.
    :param loop: Optional event loop to set the :attr:`_loop` attribute.

    .. attribute:: _loop

        The event loop used by the client IO requests.

        The event loop is stored at this attribute so that asynchronous
        method decorators such as :func:`.task` can be used.

    .. attribute:: address

        remote server UDP address.

    .. attribute:: pool

        Asynchronous client protocol :class:`.Pool`.
    '''
    protocol_factory = EchoUdpClientProtocol

    def __init__(self, address, pool_size=5, loop=None):
        super(Echo, self).__init__(loop)
        self.address = address
        self.pool = Pool(self.create_endpoint, pool_size, self._loop)

    def create_endpoint(self):
        return self.create_datagram_endpoint(remote_addr=self.address)

    @task
    def __call__(self, message):
        '''Send a ``message`` to the server and wait for a response.

        :return: a :class:`.Future`
        '''
        protocol = yield self.pool.connect()
        with protocol:
            result = yield protocol.send(message)
            coroutine_return(result)


def server(name=None, description=None, **kwargs):
    '''Create the :class:`.UdpSocketServer`.
    '''
    name = name or 'echoudpserver'
    description = description or 'Echo Udp Server'
    return UdpSocketServer(EchoUdpServerProtocol, name=name,
                           description=description, **kwargs)


if __name__ == '__main__':  # pragma nocover
    server().start()



from sqlalchemy.engine import create_engine
from sqlalchemy.ext.declarative import declarative_base, DeclarativeMeta
from sqlalchemy.orm import sessionmaker

from pulsar.apps.data import Store, register_store
from pulsar.apps.greenio import green_task
from pulsar.utils.pep import itervalues


Base = declarative_base()


class SqlStore(Store):
    '''A pulsar :class:`.Store` based on sqlalchemy
    '''
    def _init(self, **kwargs):
        dns = self._buildurl()
        self.sql_engine = create_engine(dns, **kwargs)
        self.transaction = sessionmaker(bind=self.sql_engine)

    @green_task
    def ping(self):
        conn = self.sql_engine.connect()
        r = conn.execute('select 1;')
        result = r.fetchone()
        return result[0] == 1

    @green_task
    def create_table(self, model, remove_existing=False):
        sql_model = self.sql_model(model)
        table = sql_model.__table__
        return table.create(self.sql_engine, checkfirst=remove_existing)

    @green_task
    def drop_table(self, model, remove_existing=False):
        sql_model = self.sql_model(model)
        table = sql_model.__table__
        return table.drop(self.sql_engine, checkfirst=remove_existing)

    def close(self):
        # make sure no connections are opened
        self.sql_engine.dispose()

    # ODM SUPPORT

    def create_model(self, manager, *args, **kwargs):
        sql_model = self.sql_model(manager._model)
        return sql_model(*args, **kwargs)

    @green_task
    def execute_transaction(self, transaction):
        '''Execute a ``transaction``
        '''
        transaction.commit()
        return list(itervalues(transaction.identity_map))
    # INTERNALS

    def sql_model(self, model):
        sql_model = getattr(model, '_sql_model', None)
        if not sql_model:
            meta = model._meta
            fields = {'__tablename__': meta.table_name}
            for field in model._meta.dfields.values():
                fields[field.store_name] = field.sql_alchemy_column()
            sql_model = DeclarativeMeta(meta.name, (Base,), fields)
            model._sql_model = sql_model
            sql_model._meta = meta
        return sql_model



import os
import sys
import glob
from importlib import *


def expand_star(mod_name):
    """Expand something like 'unuk.tasks.*' into a list of all the modules
    there.
    """
    expanded = []
    mod_dir = os.path.dirname(
        __import__(mod_name[:-2], {}, {}, ['']).__file__)
    for f in glob.glob1(mod_dir, "[!_]*.py"):
        expanded.append('%s.%s' % (mod_name[:-2], f[:-3]))
    return expanded


def import_modules(modules):
    '''Safely import a list of *modules*
    '''
    all = []
    for mname in modules:
        if mname.endswith('.*'):
            to_load = expand_star(mname)
        else:
            to_load = [mname]
        for module in to_load:
            try:
                all.append(import_module(module))
            except ImportError as e:
                pass
    return all


def module_attribute(dotpath, default=None, safe=False):
    '''Load an attribute from a module.

    If the module or the attribute is not available, return the default
    argument if *safe* is `True`.
    '''
    if dotpath:
        bits = str(dotpath).split('.')
        try:
            module = import_module('.'.join(bits[:-1]))
            return getattr(module, bits[-1], default)
        except Exception as e:
            if not safe:
                raise
            return default
    else:
        if not safe:
            raise ImportError()
        return default


def py_file(name):
    if name.endswith('.py'):
        return name[:-3]
    elif name.endswith('.pyc'):
        return name[:-4]
    else:
        return name


def import_system_file(mod, add_to_path=True):
    if os.path.isfile(mod):
        # it is a file in the system path
        dir, name = os.path.split(mod)
        names = [py_file(name)]
        while dir and dir not in sys.path:
            ndir, name = os.path.split(dir)
            if dir == ndir:
                dir = ''
                break
            dir = ndir
            names.insert(0, name)
        # the file was not in the system path
        if not dir and add_to_path:
            dir, name = os.path.split(mod)
            if dir and dir != mod:
                sys.path.append(dir)
            mod_name = py_file(name)
        else:
            mod_name = '.'.join(names)
        return import_module(mod_name)



import unittest

from pulsar import send

from .manage import DiningPhilosophers


class TestPhylosophers(unittest.TestCase):
    app_cfg = None
    concurrency = 'thread'

    @classmethod
    def setUpClass(cls):
        app = DiningPhilosophers(name='plato',
                                 concurrency=cls.concurrency)
        cls.app_cfg = yield send('arbiter', 'run', app)

    def test_info(self):
        while True:
            philo = []
            while len(philo) < 5:
                info = yield send('plato', 'info')
                philo = info.get('workers', [])
            all = []
            for data in philo:
                p = data.get('philosopher')
                if p:
                    all.append(p)
            if len(all) == 5:
                break

    @classmethod
    def tearDownClass(cls):
        if cls.app_cfg is not None:
            return send('arbiter', 'kill_actor', cls.app_cfg.name)



'''The :class:`TestSuite` is a testing framework for both synchronous and
asynchronous applications and for running tests in parallel
on multiple threads or processes.
It is used for testing pulsar but it can be used
as a test suite for any other library.

.. _apps-test-intro:

Introduction
====================

To get started with pulsar asynchronous test suite is easy. The first thing to
do is to create a python ``script`` on the top level directory of your library,
let's call it ``runtests.py``::

    from pulsar.apps import TestSuite

    if __name__ == '__main__':
        TestSuite(description='Test suite for my library',
                  modules=('regression',
                           ('examples', 'tests'))).start()

where ``modules`` is a tuple/list which specifies where to
:ref:`search for test cases <apps-test-loading>` to run.
In the above example the test suite will look for all python files
in the ``regression`` module (in a recursive fashion), and for modules
called ``tests`` in the ``examples`` module.

To run the test suite::

    python runtests.py

Type::

    python runtests.py --help

For a list different options/parameters which can be used when running tests.

.. note::

    Pulsar test suite help you speed up your tests!
    Use ``setUpClass`` rather than ``setUp``, use asynchronous tests and if
    not possible add fire power with more test workers.

    Try with the ``-w 8`` command line input for a laugh.

The next step is to actually write the tests.

Writing a Test Case
===========================
Only subclasses of  :class:`~unittest.TestCase` are collected by this
application.
When running a test, pulsar looks for two extra method: ``_pre_setup`` and
``_post_teardown``. If the former is available, it is run just before the
:meth:`~unittest.TestCase.setUp` method while if the latter is available,
it is run just after the :meth:`~unittest.TestCase.tearDown` method.
In addition, if the :meth:`~unittest.TestCase.setUpClass`
class methods is available, it is run just before
all tests functions are run and the :meth:`~unittest.TestCase.tearDownClass`,
if available, is run just after all tests functions are run.

An example test case::

    import unittest

    class MyTest(unittest.TestCase):

        def test_async_test(self):
            result = yield maybe_async_function()
            self.assertEqual(result, ...)

        def test_simple_test(self):
            self.assertEqual(1, 1)

.. note::

    Test functions are asynchronous, when they return a generator or a
    :class:`~asyncio.Future`, synchronous, when they return anything else.

.. _apps-test-loading:

Loading Tests
=================

The loading of test cases is controlled by the ``modules`` parameter when
initialising the :class:`TestSuite`::

    from pulsar.apps import TestSuite

    if __name__ == '__main__':
        TestSuite(modules=('tests',
                          ('examples','tests'))).start()

The :class:`TestSuite` loads tests via the :class:`~loader.TestLoader` class.

In the context of explaining the rules for loading tests, we refer to
an ``object`` as

* a ``module`` (including a directory module)
* a python ``class``.

with this in mind:

* Directories that aren't packages are not inspected. This means that if a
  directory does not have a ``__init__.py`` file, it won't be inspected.
* Any class that is a :class:`~unittest.TestCase` subclass is collected.
* If an ``object`` starts with ``_`` or ``.`` it won't be collected,
  nor will any ``objects`` it contains.
* If an ``object`` defines a ``__test__`` attribute that does not evaluate to
  ``True``, that object will not be collected, nor will any ``objects``
  it contains.

The ``modules`` parameter is a tuple or list of entries where each single
``entry`` follows the rules:

* If the ``entry`` is a string, it indicates the **dotted path** relative
  to the **root** directory (the directory of the script running the tests).
  For example::

      modules = ['tests', 'moretests.here']

* If an entry is two elements tuple, than the first element represents
  a **dotted path** relative to the **root** directory and the
  second element a **pattern** which modules (files or directories) must match
  in order to be included in the search. For example::

      modules = [...,
                 ('examples', 'test*')
                 ]

  load modules, from inside the ``examples`` module, starting with ``test``.

* If an entry is a three elements tuple, it is the same as the *two elements
  tuple* rule with the third element which specifies the top level **tag**
  for all tests in this entry. For example::

      modules = [...,
                  ('bla', '*', 'foo')
                ]

For example, the following::

    modules = ['test', ('bla', '*', 'foo'), ('examples','test*')]

loads

* all tests from modules in the ``test`` directory,
* all tests from the ``bla`` directory with top level tag ``foo``,
* all tests from modules which starts with *test* from the ``examples``
  directory.

All top level modules will be added to the python ``path``.


.. _test-suite-options:

Options
==================

All standard :ref:`settings <settings>` can be applied to the test application.
In addition, the following options are
:ref:`test-suite specific <setting-section-test>`:

.. _apps-test-sequential:

sequential
~~~~~~~~~~~~~~~~~~~
By default, test functions within a :class:`~unittest.TestCase`
are run in asynchronous fashion. This means that several test functions
may be executed at once depending on their return values.
By specifying the :ref:`--sequential <setting-sequential>` command line option,
the :class:`TestSuite`
forces tests to be run in a sequential model, one after the other::

    python runtests.py --sequential

Alternatively, if you need to specify a Testcase which always runs its test
functions in a sequential way, you can use the :func:`.sequential` decorator::

    from pulsar.apps.test import unittest, sequential

    @sequential
    class MyTestCase(unittest.TestCase):
        ...


list labels
~~~~~~~~~~~~~~~~~~~
By passing the ``-l`` or :ref:`--list-labels <setting-list_labels>` flag
to the command line, the full list of test labels available is displayed::

    python runtests.py -l


test timeout
~~~~~~~~~~~~~~~~~~~
When running asynchronous tests, it can be useful to set a cap on how
long a test function can wait for results. This is what the
:ref:`--test-timeout <setting-test_timeout>` command line flag does::

    python runtests.py --test-timeout 10

Set the test timeout to 10 seconds.


Test Plugins
==================
A :class:`.TestPlugin` is a way to extend the test suite with additional
:ref:`options <test-suite-options>` and behaviours implemented in
the various plugin's callbacks.
There are two basic rules for plugins:

* Test plugin classes should subclass :class:`.TestPlugin`.
* Test plugins may implement any of the methods described in the class
  :class:`.Plugin` interface.

Pulsar ships with two battery-included plugins:

.. _bench-plugin:

Benchmark
~~~~~~~~~~~~~

.. automodule:: pulsar.apps.test.plugins.bench


.. _profile-plugin:

Profile
~~~~~~~~~~~~~

.. automodule:: pulsar.apps.test.plugins.profile


API
=========

Test Suite
~~~~~~~~~~~~~~~~~~~~~~~

.. autoclass:: pulsar.apps.test.TestSuite
   :members:
   :member-order: bysource


Test Loader
~~~~~~~~~~~~~~~~~~~~~~~

.. automodule:: pulsar.apps.test.loader


Plugin
~~~~~~~~~~~~~~~~~~~

.. autoclass:: pulsar.apps.test.result.Plugin
   :members:
   :member-order: bysource


Test Runner
~~~~~~~~~~~~~~~~~~~

.. autoclass:: pulsar.apps.test.result.TestRunner
   :members:
   :member-order: bysource


Test Result
~~~~~~~~~~~~~~~~~~~

.. autoclass:: pulsar.apps.test.result.TestResult
   :members:
   :member-order: bysource


Test Plugin
~~~~~~~~~~~~~~~~~~~

.. autoclass:: pulsar.apps.test.plugins.base.TestPlugin
   :members:
   :member-order: bysource

Populate
~~~~~~~~~~~~~

A useful function for populating random data::

    from pulsar.apps.test import populate

    data = populate('string', 100)

gives you a list of 100 random strings


.. autofunction:: pulsar.apps.test.populate.populate


Utilities
================

.. automodule:: pulsar.apps.test.utils

'''
import sys
import unittest
from functools import partial

import pulsar
from pulsar import multi_async
from pulsar.apps import tasks
from pulsar.apps.data import create_store
from pulsar.apps.ds import PulsarDS
from pulsar.utils.log import lazyproperty
from pulsar.utils.config import section_docs, TestOption
from pulsar.utils.pep import default_timer, to_string

from .case import mock
from .populate import populate
from .result import *
from .plugins.base import *
from .loader import *
from .utils import *
from .wsgi import *
from .pep import pep8_run


pyver = '%s.%s' % (sys.version_info[:2])

section_docs['Test'] = '''
This section covers configuration parameters used by the
:ref:`Asynchronous/Parallel test suite <apps-test>`.'''


class ExitTest(Exception):
    pass


class TestVerbosity(TestOption):
    name = 'verbosity'
    flags = ['--verbosity']
    type = int
    default = 1
    desc = """Test verbosity, 0, 1, 2, 3"""


class TestTimeout(TestOption):
    name = 'test_timeout'
    flags = ['--test-timeout']
    validator = pulsar.validate_pos_int
    type = int
    default = 30
    desc = '''\
        Tests which take longer than this many seconds are timed-out
        and failed.'''


class TestLabels(TestOption):
    name = "labels"
    nargs = '*'
    validator = pulsar.validate_list
    desc = """Optional test labels to run.

    If not provided all tests are run.
    To see available labels use the -l option.
    """


class TestExcludeLabels(TestOption):
    name = "exclude_labels"
    flags = ['-e', '--exclude-labels']
    nargs = '*'
    desc = 'Exclude a group o labels from running.'
    validator = pulsar.validate_list


class TestSize(TestOption):
    name = 'size'
    flags = ['--size']
    choices = ('tiny', 'small', 'normal', 'big', 'huge')
    default = 'normal'
    desc = """Optional test size."""


class TestList(TestOption):
    name = "list_labels"
    flags = ['-l', '--list-labels']
    action = 'store_true'
    default = False
    validator = pulsar.validate_bool
    desc = """List all test labels without performing tests."""


class TestSequential(TestOption):
    name = "sequential"
    flags = ['--sequential']
    action = 'store_true'
    default = False
    validator = pulsar.validate_bool
    desc = """Run test functions sequentially.

    Don't run them asynchronously.
    """


class TestShowLeaks(TestOption):
    name = "show_leaks"
    flags = ['--show-leaks']
    nargs = '?'
    choices = (0, 1, 2)
    const = 1
    type = int
    default = 0
    desc = """Shows memory leaks.

    Run the garbage collector before a process-based actor dies and shows
    the memory leak report.
    """


class TestPep8(TestOption):
    name = "pep8"
    flags = ['--pep8']
    nargs = '*'
    validator = pulsar.validate_list
    desc = """Run pep8"""


class TestSuite(tasks.TaskQueue):
    '''An asynchronous test suite which works like a task queue.

    Each task is a group of test methods in a python TestCase class.

    :parameter modules: An iterable over modules where to look for tests.
        Alternatively it can be a callable returning the iterable over modules.
        For example::

            suite = TestSuite(modules=('regression',
                                       ('examples','tests'),
                                       ('apps','test_*')))

            def get_modules(suite):
                ...

            suite = TestSuite(modules=get_modules)

        If not provided it is set as default to ``["tests"]`` which loads all
        python module from the tests module in a recursive fashion.
        Check the the :class:`.TestLoader` for detailed information.

    :parameter result_class: Optional class for collecting test results.
        By default it used the standard :class:`.TestResult`.
    :parameter plugins: Optional list of :class:`.TestPlugin` instances.
    '''
    name = 'test'
    cfg = pulsar.Config(apps=('tasks', 'test'),
                        loglevel=['none'],
                        task_paths=['pulsar.apps.test.case'],
                        plugins=())

    def new_runner(self):
        '''The :class:`.TestRunner` driving test cases.
        '''
        if mock is None:    # pragma    nocover
            raise ExitTest('python %s requires mock library for pulsar '
                           'test suite application' % pyver)
        result_class = getattr(self, 'result_class', None)
        stream = pulsar.get_stream(self.cfg)
        runner = TestRunner(self.cfg.plugins, stream, result_class)
        abort_message = runner.configure(self.cfg)
        if abort_message:    # pragma    nocover
            raise ExitTest(str(abort_message))
        self.runner = runner
        return runner

    @lazyproperty
    def loader(self):
        # When config is available load the tests and check what type of
        # action is required.
        modules = self.cfg.get('modules')
        # Create a runner and configure it
        runner = self.new_runner()
        if not modules:
            modules = ['tests']
        if hasattr(modules, '__call__'):
            modules = modules(self)
        return TestLoader(self.root_dir, modules, runner, logger=self.logger)

    def on_config(self, arbiter):
        stream = arbiter.stream
        try:
            loader = self.loader
        except ExitTest as e:
            stream.writeln(str(e))
            return False
        stream = arbiter.stream
        stream.writeln(sys.version)
        if self.cfg.list_labels:    # pragma    nocover
            tags = self.cfg.labels
            if tags:
                s = '' if len(tags) == 1 else 's'
                stream.writeln('\nTest labels for%s %s:' %
                               (s, ', '.join(tags)))
            else:
                stream.writeln('\nAll test labels:')
            stream.writeln('')

            def _tags():
                for tag, mod in loader.testmodules(tags):
                    doc = mod.__doc__
                    if doc:
                        tag = '{0} - {1}'.format(tag, doc)
                    yield tag
            for tag in sorted(_tags()):
                stream.writeln(tag)
            stream.writeln('')
            return False
        elif self.cfg.pep8:
            msg, code = pep8_run(self.cfg.pep8)
            stream.writeln(msg)
            if code:
                sys.exit(code)
            return False

    def monitor_start(self, monitor):
        '''When the monitor starts load all test classes into the queue'''
        # Create a datastore for this test suite
        if not self.cfg.task_backend:
            server = PulsarDS(bind='127.0.0.1:0', workers=0,
                              key_value_save=[],
                              name='%s_store' % self.name)
            yield server()
            address = 'pulsar://%s:%s' % server.cfg.addresses[0]
        else:
            address = self.cfg.task_backend

        store = create_store(address, pool_size=2, loop=monitor._loop)
        self.get_backend(store)
        loader = self.loader
        tags = self.cfg.labels
        exclude_tags = self.cfg.exclude_labels
        if self.cfg.show_leaks:
            show = show_leaks if self.cfg.show_leaks == 1 else hide_leaks
            self.cfg.set('when_exit', show)
            arbiter = pulsar.arbiter()
            arbiter.cfg.set('when_exit', show)
        try:
            tests = []
            loader.runner.on_start()
            for tag, testcls in loader.testclasses(tags, exclude_tags):
                suite = loader.runner.loadTestsFromTestCase(testcls)
                if suite and suite._tests:
                    tests.append((tag, testcls))
            self._time_start = None
            if tests:
                self.logger.info('loading %s test classes', len(tests))
                monitor.cfg.set('workers', min(self.cfg.workers, len(tests)))
                self._time_start = default_timer()
                queued = []
                self._tests_done = set()
                self._tests_queued = None
                #
                # Bind to the task_done event
                self.backend.bind_event('task_done',
                                        partial(self._test_done, monitor))
                for tag, testcls in tests:
                    r = self.backend.queue_task('test', testcls=testcls,
                                                tag=tag)
                    queued.append(r)
                queued = yield multi_async(queued)
                self.logger.debug('loaded %s test classes', len(tests))
                self._tests_queued = set(queued)
                yield self._test_done(monitor)
            else:   # pragma    nocover
                raise ExitTest('Could not find any tests.')
        except ExitTest as e:   # pragma    nocover
            monitor.stream.writeln(str(e))
            monitor.arbiter.stop()
        except Exception:   # pragma    nocover
            monitor.logger.critical('Error occurred while starting tests',
                                    exc_info=True)
            monitor._loop.call_soon(self._exit, 3)

    @classmethod
    def create_config(cls, *args, **kwargs):
        cfg = super(TestSuite, cls).create_config(*args, **kwargs)
        if cfg.params.get('plugins') is None:
            cfg.params['plugins'] = ()
        for plugin in cfg.params['plugins']:
            cfg.settings.update(plugin.config.settings)
        return cfg

    def arbiter_params(self):
        params = super(TestSuite, self).arbiter_params()
        params['concurrency'] = self.cfg.concurrency
        return params

    def _test_done(self, monitor, task_id=None, exc=None):
        runner = self.runner
        if task_id:
            self._tests_done.add(to_string(task_id))
        if self._tests_queued is not None:
            left = self._tests_queued.difference(self._tests_done)
            if not left:
                tests = yield self.backend.get_tasks(self._tests_done)
                self.logger.info('All tests have finished.')
                time_taken = default_timer() - self._time_start
                for task in tests:
                    runner.add(task.get('result'))
                runner.on_end()
                runner.printSummary(time_taken)
                # Shut down the arbiter
                if runner.result.errors or runner.result.failures:
                    exit_code = 2
                else:
                    exit_code = 0
                monitor._loop.call_soon(self._exit, exit_code)

    def _exit(self, exit_code):
        raise pulsar.HaltServer(exit_code=exit_code)



import pulsar
from pulsar.utils.httpurl import iteritems
from pulsar.apps.test.result import Plugin


__all__ = ['WrapTest', 'TestPlugin']


def as_test_setting(setting):
    setting.app = 'test'
    setting.section = "Test"
    return setting


class WrapTest(object):
    '''Wrap an underlying test case'''
    def __init__(self, test):
        self.test = test
        setattr(self, test._testMethodName, self._call)
        self.testMethod = getattr(test, test._testMethodName)

    def __str__(self):
        return self.test._testMethodName
    __repr__ = __str__

    @property
    def original_test(self):
        if isinstance(self.test, WrapTest):
            return self.test.original_test
        else:
            return self.test

    def set_test_attribute(self, name, value):
        setattr(self.original_test, name, value)

    def __getattr__(self, name):
        return getattr(self.original_test, name)

    def _call(self):
        # This is the actual function to implement
        return self.testMethod()


class TestPluginMeta(type):

    def __new__(cls, name, bases, attrs):
        settings = {}
        for base in bases:
            if isinstance(base, TestPluginMeta):
                settings.update(base.config.settings)
        for key, setting in list(iteritems(attrs)):
            if isinstance(setting, pulsar.Setting):
                attrs.pop(key)
                setting.name = setting.name or key.lower()
                settings[setting.name] = as_test_setting(setting)
        if not attrs.pop('virtual', False):
            setting_name = attrs.pop('name', name).lower()
            if setting_name:
                def_flag = '--%s' % setting_name.replace(
                    ' ', '-').replace('_', '-')
                action = attrs.pop('action', None)
                type = attrs.pop('type', None)
                default = attrs.pop('default', None)
                validator = attrs.pop('validator', None)
                nargs = attrs.pop('nargs', None)
                if (validator is None and default is None and type is None
                        and nargs is None):
                    if action is None or action == 'store_true':
                        action = 'store_true'
                        default = False
                        validator = pulsar.validate_bool
                    elif action == 'store_false':
                        default = True
                        validator = pulsar.validate_bool
                setting = pulsar.Setting(name=setting_name,
                                         desc=attrs.pop('desc', name),
                                         type=type,
                                         flags=attrs.pop('flags', [def_flag]),
                                         action=action,
                                         default=default,
                                         validator=validator,
                                         nargs=nargs)
                settings[setting.name] = as_test_setting(setting)
        attrs['config'] = pulsar.Config(settings=settings)
        return super(TestPluginMeta, cls).__new__(cls, name, bases, attrs)


class TestPlugin(TestPluginMeta('TestPluginBase',
                                (Plugin,), {'virtual': True})):
    '''Base class for :class:`.Plugin` which can be added to a
    :class:`.TestSuite` to extend its functionalities.

    If the class attribute :attr:`name` is not specified or its value validate
    as ``True``, an additional :ref:`setting <settings>` is added to the
    configuration.
    In addition, a :class:`TestPlugin` can specify several additional
    :ref:`settings <settings>` as class attributes. For example, the
    :ref:`benchmark plugin <bench-plugin>` has an additional setting
    for controlling the number of repetitions::

        class Bench(TestPlugin):
            repeat = pulsar.Setting(type=int,
                                default=1,
                                validator=pulsar.validate_pos_int,
                                desc="Default number of repetition")


    .. attribute:: name

        Class attribute used for adding the default plugin
        :ref:`setting <settings>`
        to the configuration container of the test suite application.
        If the attribute is not set, the class name in lower case is used.
        If set and validate as not ``True``, no new :ref:`setting <settings>`
        is added to the test suite configuration parameters. For example::

            class MyPlugin(TestPlugin):
                name = None

        won't add the default plugin :ref:`setting <settings>`.

    .. attribute:: desc

        Class attribute used as the description of the
        :ref:`setting <settings>`.
        If :attr:`name` is disabled, this attribute is not relevant.

    .. attribute:: config

        A :class:`.Config` container created by the :class:`TestPlugin`
        metaclass. It collects the default setting, if
        available, and any additional :ref:`settings <settings>` specified
        as class attributes.
    '''
    virtual = True

    def __new__(cls):
        o = super(TestPlugin, cls).__new__(cls)
        o.config = cls.config.copy()
        return o

    def configure(self, cfg):
        self.config = cfg



'''Tests the websocket middleware in pulsar.apps.ws.'''
import unittest

from pulsar import asyncio, send
from pulsar.apps import rpc, http, ws
from pulsar.apps.test import dont_run_with_thread
from pulsar.utils.httpurl import HTTPError
from pulsar.utils.system import json

from .manage import server


class Message(ws.WS):

    def __init__(self, loop):
        self.queue = asyncio.Queue(loop=loop)

    def get(self):
        return self.queue.get()

    def on_message(self, websocket, message):
        self.queue.put_nowait(message)


class TestWebChat(unittest.TestCase):
    app_cfg = None
    concurrency = 'thread'

    @classmethod
    def setUpClass(cls):
        s = server(bind='127.0.0.1:0', name=cls.__name__.lower(),
                   concurrency=cls.concurrency)
        cls.app_cfg = yield send('arbiter', 'run', s)
        cls.uri = 'http://%s:%s' % cls.app_cfg.addresses[0]
        cls.ws = 'ws://%s:%s/message' % cls.app_cfg.addresses[0]
        cls.rpc = rpc.JsonProxy('%s/rpc' % cls.uri)
        cls.http = http.HttpClient()

    @classmethod
    def tearDownClass(cls):
        if cls.app_cfg is not None:
            yield send('arbiter', 'kill_actor', cls.app_cfg.name)

    def test_home(self):
        response = yield self.http.get(self.uri)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.headers['content-type'],
                         'text/html; charset=utf-8')

    def test_handshake(self):
        ws = yield self.http.get(self.ws)
        response = ws.handshake
        self.assertEqual(ws.status_code, 101)
        self.assertEqual(ws.headers['upgrade'], 'websocket')
        self.assertEqual(response.connection, ws.connection)
        self.assertTrue(ws.connection)
        #
        # The connection should not be in the connection pool
        pool = self.http.connection_pools.get(ws._request.key)
        self.assertIsInstance(pool, self.http.connection_pool)
        self.assertFalse(ws.connection in pool)

    def test_rpc(self):
        '''Send a message to the rpc'''
        loop = self.http._loop
        ws = yield self.http.get(self.ws, websocket_handler=Message(loop))
        self.assertEqual(ws.status_code, 101)
        ws.write('Hello there!')
        data = yield ws.handler.get()
        data = json.loads(data)
        self.assertEqual(data['message'], 'Hello there!')
        result = yield self.rpc.message('Hi!')
        self.assertEqual(result, 'OK')
        data = yield ws.handler.get()
        data = json.loads(data)
        self.assertEqual(data['message'], 'Hi!')

    def test_invalid_method(self):
        p = rpc.JsonProxy(self.uri)
        try:
            yield p.message('ciao')
        except HTTPError as e:
            self.assertEqual(e.code, 405)
        else:
            assert False, '405 not raised'


@dont_run_with_thread
class TestWebChatProcess(TestWebChat):
    concurrency = 'process'



from .skiplist import Skiplist
from ..pep import iteritems, zip


class Zset(object):
    '''Ordered-set equivalent of redis zset.
    '''
    def __init__(self, data=None):
        self._sl = Skiplist()
        self._dict = {}
        if data:
            self.update(data)

    def __repr__(self):
        return repr(self._sl)
    __str__ = __repr__

    def __len__(self):
        return len(self._dict)

    def __iter__(self):
        for _, value in self._sl:
            yield value

    def __getstate__(self):
        return self._dict

    def __setstate__(self, state):
        self._dict = state
        self._sl = Skiplist(((score, member) for member, score
                             in iteritems(state)))

    def __eq__(self, other):
        if isinstance(other, Zset):
            return other._dict == self._dict
        return False

    def items(self):
        '''Iterable over ordered score, value pairs of this :class:`zset`
        '''
        return iter(self._sl)

    def range(self, start, end, scores=False):
        return self._sl.range(start, end, scores)

    def range_by_score(self, minval, maxval, include_min=True,
                       include_max=True, start=0, num=None, scores=False):
        return self._sl.range_by_score(minval, maxval, start=start,
                                       num=num, include_min=include_min,
                                       include_max=include_max,
                                       scores=scores)

    def score(self, member, default=None):
        '''The score of a given member'''
        return self._dict.get(member, default)

    def count(self, minval, maxval, include_min=True, include_max=True):
        return self._sl.count(minval, maxval, include_min, include_max)

    def add(self, score, val):
        r = 1
        if val in self._dict:
            sc = self._dict[val]
            if sc == score:
                return 0
            self.remove(val)
            r = 0
        self._dict[val] = score
        self._sl.insert(score, val)
        return r

    def update(self, score_vals):
        '''Update the :class:`zset` with an iterable over pairs of
scores and values.'''
        add = self.add
        for score, value in score_vals:
            add(score, value)

    def remove_items(self, items):
        removed = 0
        for item in items:
            score = self.remove(item)
            if score is not None:
                removed += 1
        return removed

    def remove(self, item):
        '''Remove ``item`` for the :class:`zset` it it exists.
        If found it returns the score of the item removed.
        '''
        score = self._dict.pop(item, None)
        if score is not None:
            index = self._sl.rank(score)
            assert index >= 0, 'could not find start range'
            for i, v in enumerate(self._sl.range(index)):
                if v == item:
                    assert self._sl.remove_range(index + i, index+i + 1) == 1
                    return score
            assert False, 'could not find element'

    def remove_range(self, start, end):
        '''Remove a range by score.
        '''
        return self._sl.remove_range(
            start, end, callback=lambda sc, value: self._dict.pop(value))

    def remove_range_by_score(self, minval, maxval,
                              include_min=True, include_max=True):
        '''Remove a range by score.
        '''
        return self._sl.remove_range_by_score(
            minval, maxval, include_min=include_min, include_max=include_max,
            callback=lambda sc, value: self._dict.pop(value))

    def clear(self):
        '''Clear this :class:`zset`.'''
        self._sl = skiplist()
        self._dict.clear()

    def rank(self, item):
        '''Return the rank (index) of ``item`` in this :class:`zset`.'''
        score = self._dict.get(item)
        if score is not None:
            return self._sl.rank(score)

    def flat(self):
        return self._sl.flat()

    @classmethod
    def union(cls, zsets, weights, oper):
        result = None
        for zset, weight in zip(zsets, weights):
            if result is None:
                result = cls()
                sl = result._sl
                for score, value in zset._sl:
                    result.add(score*weight, value)
            else:
                for score, value in zset._sl:
                    score *= weight
                    existing = sl.score(value)
                    if existing is not None:
                        score = oper(score, existing)
                    result.add(score, value)
        return result

    @classmethod
    def inter(cls, zsets, weights, oper):
        result = None
        values = None
        for zset, _ in zip(zsets, weights):
            if values is None:
                values = set(zset)
            else:
                values.intersection_update(zset)
        #
        for zset, weight in zip(zsets, weights):
            if result is None:
                result = cls()
                sl = result._sl
                for score, value in zset._sl:
                    if value in values:
                        result.add(score*weight, value)
            else:
                for score, value in zset._sl:
                    if value in values:
                        existing = result.score(value)
                        score = oper((score*weight, existing))
                        result.add(score, value)
        return result



import sys
import time
import string
import threading
from inspect import istraceback


ispy3k = sys.version_info >= (3, 0)

if ispy3k:
    default_timer = time.monotonic
else:   # pragma    nocover
    default_timer = time.time

try:
    pypy = True
    import __pypy__
except ImportError:
    pypy = False

if ispy3k:  # Python 3
    import pickle
    string_type = str
    ascii_letters = string.ascii_letters
    zip = zip
    map = map
    range = range
    chr = chr
    iteritems = lambda d: d.items()
    itervalues = lambda d: d.values()
    is_string = lambda s: isinstance(s, str)

    def to_bytes(s, encoding=None, errors=None):
        '''Convert *s* into bytes'''
        if not isinstance(s, bytes):
            return ('%s' % s).encode(encoding or 'utf-8', errors or 'strict')
        elif not encoding or encoding == 'utf-8':
            return s
        else:
            d = s.decode('utf-8')
            return d.encode(encoding, errors or 'strict')

    def to_string(s, encoding=None, errors='strict'):
        """Inverse of to_bytes"""
        if isinstance(s, bytes):
            return s.decode(encoding or 'utf-8', errors)
        else:
            return str(s)

    def native_str(s, encoding=None):
        if isinstance(s, bytes):
            return s.decode(encoding or 'utf-8')
        else:
            return s

    def force_native_str(s, encoding=None):
        if isinstance(s, bytes):
            return s.decode(encoding or 'utf-8')
        elif not isinstance(s, str):
            return str(s)
        else:
            return s

else:   # pragma : no cover
    from itertools import izip as zip, imap as map
    import cPickle as pickle
    string_type = unicode
    ascii_letters = string.letters
    range = xrange
    chr = unichr
    iteritems = lambda d: d.iteritems()
    itervalues = lambda d: d.itervalues()
    is_string = lambda s: isinstance(s, basestring)

    def to_bytes(s, encoding=None, errors='strict'):
        encoding = encoding or 'utf-8'
        if isinstance(s, bytes):
            if encoding != 'utf-8':
                return s.decode('utf-8', errors).encode(encoding, errors)
            else:
                return s
        else:
            return unicode(s).encode(encoding, errors)

    def to_string(s, encoding=None, errors='strict'):
        """Inverse of to_bytes"""
        if isinstance(s, bytes):
            return s.decode(encoding or 'utf-8', errors)
        else:
            return unicode(s)

    def native_str(s, encoding=None):
        if isinstance(s, unicode):
            return s.encode(encoding or 'utf-8')
        else:
            return s

    def force_native_str(s, encoding=None):
        if isinstance(s, unicode):
            return s.encode(encoding or 'utf-8')
        elif not isinstance(s, str):
            return str(s)
        else:
            return s



import unittest
from copy import deepcopy

from pulsar import format_traceback
from pulsar.utils.structures import AttributeDictionary

from .utils import TestFailure, LOGGER


__all__ = ['Plugin',
           'TestStream',
           'TestRunner',
           'TestResult',
           'Plugin']


STDOUT_LINE = '\nStdout:\n%s'
STDERR_LINE = '\nStderr:\n%s'


class Plugin(object):
    '''Interface for all classes which are part of the :class:`.TestRunner`.

    Most classes used by the test application are plugins, for
    example the :class:`.TestRunner` itself,
    the :class:`.TestResult` and the :class:`.TestPlugin`.
    '''
    result = None
    '''An optional result'''
    stream = None
    '''handle for writing text on the default output.

    Set by the :class:`.TestRunner` at runtime.
    '''
    descriptions = None

    def configure(self, cfg):
        '''Called once just after construction of a :class:`.TestRunner`
           and **before any test class is loaded**.

        This is a chance to configure the :class:`.Plugin` or global variables
        which may affect the way tests are run.
        If it returns something other than ``None`` (for example an abort
        message) it will stop the configuration of all subsequent
        plugins and quit the test.

        :parameter cfg: a :class:`.Config`.
        :return: ``None`` unless the tests runner must be stopped.
        '''
        pass

    @property
    def name(self):
        return self.__class__.__name__.lower()

    @property
    def count(self):
        return self.result.count if self.result else 0

    @property
    def testsRun(self):
        return self.result.testsRun if self.result else 0

    def on_start(self):
        '''Called by the :class:`.TestSuite` once only at startup.

        This callback is invoked once all tests are loaded but before
        the test suite starts running them.
        '''
        pass

    def on_end(self):
        '''Called by the :class:`.TestSuite` just before it stops.
        '''
        pass

    def loadTestsFromTestCase(self, testcls):
        '''Called when loading tests from the ``testcls`` class.

        Can be used to modify the number of test functions loaded.'''
        pass

    def startTestClass(self, testcls):
        '''Called just before a ``testcls`` runs its tests.
        '''
        pass

    def stopTestClass(self, testcls):
        '''Called just after a ``testcls`` has run its tests.
        '''
        pass

    def startTest(self, test):
        '''Called just before a ``test`` function is executed.

        This is run just before ``_pre_setup`` method.
        '''
        pass

    def stopTest(self, test):
        '''Called just after a ``test`` function has finished.

        This is run just after the ``_post_teardown`` method.
        '''
        pass

    def before_test_function_run(self, test, local):
        '''Can be used by plugins to manipulate the ``test``
        behaviour in the process domain where the test run.'''
        return test

    def after_test_function_run(self, test, local):
        '''Executed in the ``test`` process domain, after the ``test`` has
        finished.'''
        pass

    def addSuccess(self, test):
        '''Called when a ``test`` function succeed
        '''
        pass

    def addFailure(self, test, err):
        '''Called when a ``test`` function as a (test) failure
        '''
        pass

    def addError(self, test, err):
        '''Called when a ``test`` function as an (unexpected) error
        '''
        pass

    def addExpectedFailure(self, test, err):
        pass

    def addSkip(self, test, reason):
        pass

    def printErrors(self):
        pass

    def printSummary(self, timeTaken):
        pass

    def import_module(self, mod, parent=None):
        return mod

    def getDescription(self, test):
        doc_first_line = test.shortDescription()
        teststr = '%s.%s' % (test.tag, test)
        if self.descriptions and doc_first_line:
            return '\n'.join((teststr, doc_first_line))
        else:
            return teststr


class TestStream(Plugin):  # pragma    nocover
    '''Handle the writing of test results'''
    separator1 = '=' * 70
    separator2 = '-' * 70

    def __init__(self, stream, result, descriptions=True):
        self._handlers = {}
        self.stream = stream
        self.result = result
        self.descriptions = descriptions
        self.showAll = False
        self.dots = True

    def configure(self, cfg):
        verbosity = cfg.verbosity
        self.showAll = verbosity > 1
        self.dots = verbosity == 1

    def handler(self, name):
        return self._handlers.get(name, self.stream)

    def startTest(self, test):
        if self.showAll:
            self.head(test, 'Started')

    def head(self, test, v):
        v = self.getDescription(test) + ' ... %s\n' % v
        self.stream.write(v)
        self.stream.flush()

    def addSuccess(self, test):
        if self.showAll:
            self.head(test, 'OK')
        elif self.dots:
            self.stream.write('.')
            self.stream.flush()

    def addError(self, test, err):
        if self.showAll:
            self.head(test, 'ERROR')
        elif self.dots:
            self.stream.write('E')
            self.stream.flush()

    def addFailure(self, test, err):
        if self.showAll:
            self.head(test, "FAIL")
        elif self.dots:
            self.stream.write('F')
            self.stream.flush()

    def addSkip(self, test, reason):
        if self.showAll:
            self.head(test, "skipped {0!r}".format(reason))
        elif self.dots:
            self.stream.write("s")
            self.stream.flush()

    def addExpectedFailure(self, test, err):
        if self.showAll:
            self.head(test, "expected failure")
        elif self.dots:
            self.stream.write("x")
            self.stream.flush()

    def addUnexpectedSuccess(self, test):
        if self.showAll:
            self.head(test, "unexpected success")
        elif self.dots:
            self.stream.write("u")
            self.stream.flush()

    def printErrors(self):
        if self.dots or self.showAll:
            self.stream.writeln()
        self.printErrorList('ERROR', self.result.errors)
        self.printErrorList('FAIL', self.result.failures)
        return True

    def printErrorList(self, flavour, errors):
        for test, err in errors:
            self.stream.writeln(self.separator1)
            self.stream.writeln("%s: %s" % (flavour, test))
            self.stream.writeln(self.separator2)
            self.stream.writeln("%s" % err)

    def printSummary(self, timeTaken):
        '''Write the summuray of tests results.'''
        stream = self.stream
        result = self.result
        self.printErrors()
        run = result.testsRun
        stream.writeln("Ran %d test%s in %.3fs" %
                       (run, run != 1 and "s" or "", timeTaken))
        stream.writeln()

        expectedFails = unexpectedSuccesses = skipped = 0
        results = map(len, (result.expectedFailures,
                            result.unexpectedSuccesses,
                            result.skipped))
        expectedFails, unexpectedSuccesses, skipped = results

        infos = []
        if not result.wasSuccessful():
            stream.write("FAILED")
            failed, errored = map(len, (result.failures, result.errors))
            if failed:
                infos.append("failures=%d" % failed)
            if errored:
                infos.append("errors=%d" % errored)
        else:
            stream.write("OK")
        if skipped:
            infos.append("skipped=%d" % skipped)
        if expectedFails:
            infos.append("expected failures=%d" % expectedFails)
        if unexpectedSuccesses:
            infos.append("unexpected successes=%d" % unexpectedSuccesses)
        if infos:
            stream.writeln(" (%s)" % (", ".join(infos),))
        else:
            stream.write("\n")

        return True


class TestResult(Plugin):
    '''A :class:`.Plugin` for collecting results/failures for test runs.

    Each :class:`.Plugin` can access the :class:`.TestRunner` ``result``
    object via the :attr:`~Plugin.result` attribute.
    '''
    def __init__(self, descriptions=True):
        self.descriptions = descriptions
        self._testsRun = 0
        self._count = 0
        self.failures = []
        self.errors = []
        self.skipped = []
        self.expectedFailures = []
        self.unexpectedSuccesses = []

    @property
    def count(self):
        return self._count

    @property
    def testsRun(self):
        return self._testsRun

    @property
    def result(self):
        return self

    def startTest(self, test):
        '''Increase the test counter
        '''
        self._testsRun += 1

    def addError(self, test, err):
        '''Called when an unexpected error has occurred.

        ``err`` is a tuple of values as returned by ``sys.exc_info()``
        '''
        self._add_error(test, err, self.errors)

    def addFailure(self, test, err):
        '''Called when an test failure has occurred.

        ``err`` is a tuple of values as returned by ``sys.exc_info()``
        '''
        self._add_error(test, err, self.failures)

    def addSkip(self, test, reason):
        """Called when a test is skipped."""
        self.skipped.append((self.getDescription(test), reason))

    def addExpectedFailure(self, test, err):
        """Called when an expected failure/error occured."""
        self._add_error(test, err, self.expectedFailures)

    def addUnexpectedSuccess(self, test):
        """Called when a test was expected to fail, but succeed."""
        self.unexpectedSuccesses.append(self.getDescription(test))

    def _add_error(self, test, exc, container):
        if not isinstance(exc, TestFailure):
            exc = TestFailure(exc)
        test = self.getDescription(test)
        container.append((test, str(exc)))

    def add(self, result):
        self._count += 1
        self._testsRun += result.testsRun
        self.failures.extend(result.failures)
        self.errors.extend(result.errors)
        self.skipped.extend(result.skipped)
        self.expectedFailures.extend(result.expectedFailures)
        self.unexpectedSuccesses.extend(result.unexpectedSuccesses)

    def wasSuccessful(self):
        "Tells whether or not this result was a success"
        return len(self.failures) == len(self.errors) == 0


def testsafe(name, return_val=None):
    if not return_val:
        return_val = lambda c: None

    def _(self, *args):
        for p in self.plugins:
            try:
                c = getattr(p, name)(*args)
                if c is not None:
                    return return_val(c)
            except Exception:
                LOGGER.exception('Unhadled error in %s.%s' % (p, name))
    return _


class TestRunner(Plugin):
    '''A :class:`.Plugin` for asynchronously running tests.
    '''
    def __init__(self, plugins, stream, writercls=None, descriptions=True,
                 logger=None):
        self.descriptions = descriptions
        self.plugins = []
        writercls = writercls or TestStream
        result = TestResult(descriptions=self.descriptions)
        stream = writercls(stream, result, descriptions=self.descriptions)
        for p in plugins:
            p = deepcopy(p)
            p.descriptions = self.descriptions
            p.result = result
            p.stream = stream
            self.plugins.append(p)
        self.plugins.append(result)
        self.plugins.append(stream)
        self.stream = stream
        self.result = result
        self.loader = unittest.TestLoader()

    configure = testsafe('configure', lambda c: c)
    on_start = testsafe('on_start')
    on_end = testsafe('on_end')
    startTestClass = testsafe('startTestClass')
    stopTestClass = testsafe('stopTestClass')
    startTest = testsafe('startTest')
    stopTest = testsafe('stopTest')
    addSuccess = testsafe('addSuccess')
    addFailure = testsafe('addFailure')
    addExpectedFailure = testsafe('addExpectedFailure')
    addError = testsafe('addError')
    addSkip = testsafe('addSkip')
    printErrors = testsafe('printErrors')
    printSummary = testsafe('printSummary')

    def loadTestsFromTestCase(self, test_cls):
        '''Load all ``test`` functions for the ``test_cls``
        '''
        c = testsafe('loadTestsFromTestCase', lambda v: v)(self, test_cls)
        if c is None:
            return self.loader.loadTestsFromTestCase(test_cls)
        else:
            return c

    def add(self, result):
        '''Add a new result to the :attr:`~.Plugin.result` container
        '''
        self.result.add(result)

    def import_module(self, mod, parent=None):
        for p in self.plugins:
            mod = p.import_module(mod, parent)
            if not mod:
                return
        return mod

    def before_test_function_run(self, test):
        '''Called just before the test is run,
        in the test process domain.'''
        test.plugins = plugins = {}
        for p in self.plugins:
            local = AttributeDictionary()
            plugins[p.name] = local
            test = p.before_test_function_run(test, local) or test
        return test

    def after_test_function_run(self, test):
        '''Called just after the test has finished,
        in the test process domain.'''
        for p in self.plugins:
            local = test.plugins.get(p.name)
            if local is not None:
                p.after_test_function_run(test, local)



'''Operative system specific functions and classes.
'''
import os

from .runtime import Platform

platform = Platform()
seconds = platform.seconds

from .base import *

if platform.type == 'win':    # pragma nocover
    from .windowssystem import *
elif platform.is_appengine:
    from .appengine import *
else:
    from .posixsystem import *


try:
    import psutil
except ImportError:    # pragma    nocover
    psutil = None

try:
    import ujson as json
except ImportError:     # pragma    nocover
    import json


memory_symbols = ('K', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y')
memory_size = dict(((s, 1 << (i+1)*10) for i, s in enumerate(memory_symbols)))


def convert_bytes(b):
    '''Convert a number of bytes into a human readable memory usage, bytes,
kilo, mega, giga, tera, peta, exa, zetta, yotta'''
    if b is None:
        return '#NA'
    for s in reversed(memory_symbols):
        if b >= memory_size[s]:
            value = float(b) / memory_size[s]
            return '%.1f%sB' % (value, s)
    return "%sB" % b


def process_info(pid=None):
    '''Returns a dictionary of system information for the process ``pid``.

    It uses the psutil_ module for the purpose. If psutil_ is not available
    it returns an empty dictionary.

    .. _psutil: http://code.google.com/p/psutil/
    '''
    if psutil is None:  # pragma    nocover
        return {}
    pid = pid or os.getpid()
    try:
        p = psutil.Process(pid)
    # this fails on platforms which don't allow multiprocessing
    except psutil.NoSuchProcess:  # pragma    nocover
        return {}
    else:
        mem = p.memory_info()
        return {'memory': mem.rss,
                'memory_virtual': mem.vms,
                'cpu_percent': p.cpu_percent(),
                'nice': p.nice(),
                'num_threads': p.num_threads()}



from pulsar import CommandNotFound
from pulsar.utils.pep import default_timer

from .futures import Future, chain_future
from .consts import *

__all__ = ['ActorProxy',
           'ActorProxyMonitor',
           'get_proxy',
           'command',
           'get_command']

global_commands_table = {}


def get_proxy(obj, safe=False):
    if isinstance(obj, ActorProxy):
        return obj
    elif hasattr(obj, 'proxy'):
        return get_proxy(obj.proxy)
    else:
        if safe:
            return None
        else:
            raise ValueError('"%s" is not an actor or actor proxy.' % obj)


def actorid(actor):
    return actor.identity() if hasattr(actor, 'identity') else actor


def get_command(name):
    '''Get the command function *name*'''
    command = global_commands_table.get(name.lower())
    if not command:
        raise CommandNotFound(name)
    return command


class command:
    '''Decorator for pulsar command functions.

    :parameter ack: ``True`` if the command acknowledge the sender with a
        response. Usually is set to ``True`` (which is also the default value).
    '''
    def __init__(self, ack=True):
        self.ack = ack

    def __call__(self, f):
        self.name = f.__name__.lower()

        def command_function(request, args, kwargs):
            return f(request, *args, **kwargs)

        command_function.ack = self.ack
        command_function.__name__ = self.name
        command_function.__doc__ = f.__doc__
        global_commands_table[self.name] = command_function
        return command_function


class ActorIdentity(object):

    def identity(self):
        return self.aid


def actor_proxy_future(aid, future=None):
    self = ActorProxyFuture()
    if isinstance(aid, ActorProxyMonitor):
        assert future is None
        aid.callback = self
        self.aid = aid.aid
    else:
        self.aid = aid
        chain_future(future, next=self)
    return self


class ActorProxyFuture(Future, ActorIdentity):
    '''A :class:`.Future` for an :class:`.ActorProxy`.

    The callback will be an :class:`.ActorProxy` which will be received once
    the remote :class:`.Actor` is fully functional.

    .. attribute:: aid

        The the remote :attr:`.Actor` id

    '''
    def __repr__(self):
        if self.done():
            return '%s(%s)' % (self.__class__.__name__, self.aid)
        else:
            return '%s(%s) PENDING' % (self.__class__.__name__, self.aid)
    __str__ = __repr__


class ActorProxy(ActorIdentity):
    '''A proxy for a remote :class:`.Actor`.

    This is a lightweight class which delegates function calls to the
    underlying remote object.

    It is picklable and therefore can be send from actor to actor using
    :ref:`actor message passing <tutorials-messages>`.

    For example, lets say we have a proxy ``a``, to send a message to it::

        from pulsar import send

        send(a, 'echo', 'hello there!')

    will send the :ref:`command <actor_commands>` ``echo`` to actor ``a`` with
    parameter ``"hello there!"``.

    .. attribute:: aid

        Unique ID for the remote :class:`.Actor`

    .. attribute:: address

        the socket address of the underlying :attr:`.Actor.mailbox`.

    '''
    def __init__(self, impl):
        self.aid = impl.aid
        self.name = impl.name
        self.cfg = impl.cfg
        self.address = getattr(impl, 'address', None)

    def __repr__(self):
        return '%s(%s)' % (self.name, self.aid)
    __str__ = __repr__

    @property
    def proxy(self):
        return self

    def __eq__(self, o):
        o = get_proxy(o, True)
        return o and self.aid == o.aid

    def __ne__(self, o):
        return not self.__eq__(o)


class ActorProxyMonitor(ActorProxy):
    '''A specialised :class:`.ActorProxy` class.

    It contains additional information about the remote underlying
    :class:`.Actor`. Instances of this class serialise into
    :class:`.ActorProxy`.

    The :class:`.ActorProxyMonitor` is special since it lives in the
    :class:`.Arbiter` domain and it is used by the :class:`.Arbiter`
    (or a :class:`.Monitor`) to monitor the state of the spawned actor.

    .. attribute:: impl

        The :class:`.Concurrency` instance for the remote :class:`.Actor`. This
        dictionary is constantly updated by the remote actor by sending the
        :ref:`info message <actor_info_command>`.

    .. attribute:: info

        Dictionary of information regarding the remote :class:`.Actor`

    .. attribute:: mailbox

        This is the connection with the remote actor. It is available once the
        :ref:`actor handshake <handshake>` between the actor and the monitor
        has completed. The :attr:`mailbox` is a server-side
        :class:`.MailboxProtocol` instance and it is used
        by the :func:`.send` function to send messages to the remote actor.
    '''
    monitor = None

    def __init__(self, impl):
        self.impl = impl
        self.info = {}
        self.mailbox = None
        self.callback = None
        self.spawning_start = None
        self.stopping_start = None
        super(ActorProxyMonitor, self).__init__(impl)

    @property
    def notified(self):
        '''Last time this :class:`.ActorProxyMonitor` was notified by the
        remote actor.'''
        return self.info.get('last_notified')

    @property
    def pid(self):
        return self.impl.pid

    @property
    def proxy(self):
        '''The :class:`.ActorProxy` for this monitor.'''
        return ActorProxy(self)

    def __reduce__(self):
        return self.proxy.__reduce__()

    def is_alive(self):
        '''``True`` if underlying actor is alive.
        '''
        return self.impl.is_alive()

    def terminate(self):
        '''Terminate life of underlying actor.
        '''
        self.impl.terminate()

    def join(self, timeout=None):
        '''Wait until the underlying actor terminates.

        If ``timeout`` is provided, it raises an exception if the timeout
        is reached.
        '''
        self.impl.join(timeout=timeout)

    def start(self):
        '''Start the remote actor.
        '''
        self.spawning_start = default_timer()
        self.impl.start()

    def should_be_alive(self):
        if not self.mailbox:
            return default_timer() - self.spawning_start > ACTOR_ACTION_TIMEOUT
        else:
            return True

    def should_terminate(self):
        if self.stopping_start is None:
            self.stopping_start = default_timer()
            return False
        else:
            dt = default_timer() - self.stopping_start
            return dt if dt >= ACTOR_ACTION_TIMEOUT else False



import os
import signal
import subprocess

from .base import *

__all__ = ['close_on_exec',
           'Waker',
           'daemonize',
           'socketpair',
           'EXIT_SIGNALS',
           'get_uid',
           'get_gid',
           'get_maxfd',
           'set_owner_process',
           'current_process']

# standard signal quit
EXIT_SIGNALS = (signal.SIGINT, signal.SIGTERM, signal.SIGABRT, signal.SIGQUIT)
# Default maximum for the number of available file descriptors.
REDIRECT_TO = getattr(os, "devnull", "/dev/null")

socketpair = None


def get_parent_id():
    return os.getppid()


def chown(path, uid, gid):
    pass


def close_on_exec(fd):
    pass


def get_uid(user=None):
    if not user:
        return os.geteuid()
    elif user.isdigit() or isinstance(user, int):
        return int(user)
    else:
        return pwd.getpwnam(user).pw_uid


def get_gid(group=None):
    if not group:
        return os.getegid()
    elif group.isdigit() or isinstance(group, int):
        return int(group)
    else:
        return grp.getgrnam(group).gr_gid


def set_owner_process(uid, gid):
    pass


def setpgrp():
    os.setpgrp()


def get_maxfd():
    return MAXFD


def daemonize():    # pragma    nocover
    pass


class MockProcess:
    pass


_process = MockProcess()


def current_process():
    return _process


class Waker(object):

    def fileno(self):
        pass

    def wake(self):
        pass



import zlib

from pulsar.utils import websocket

############################################################################
#  x-webkit-deflate-frame     Extension
#
# http://code.google.com/p/pywebsocket/


class deflate_frame(websocket.Extension):

    def __init__(self, window_bits=None):
        self.window_bits = window_bits or zlib.MAX_WBITS

    def receive(self, frame, application_data):
        if frame.rsv1 == 1:
            application_data += b'\x00\x00\xff\xff'
            return zlib.decompress(application_data)
        else:
            return application_data

    def send(self, frame, application_data):
        application_data = self._compress.compress(application_data)
        return application_data


# websocket.WS_EXTENSIONS['x-webkit-deflate-frame'] = deflate_frame



import logging
import threading
import weakref
from multiprocessing import dummy, current_process
from functools import partial

try:
    import queue
except ImportError:  # pragma nocover
    import Queue as queue
ThreadQueue = queue.Queue
Empty = queue.Empty
Full = queue.Full

from .access import (asyncio, selectors, new_event_loop, get_actor, set_actor,
                     events, thread_data, _StopError, BaseEventLoop)
from .futures import Future, Task, async, AsyncObject
from .consts import ACTOR_STATES


__all__ = ['Thread', 'IOqueue', 'ThreadPool', 'ThreadQueue', 'Empty', 'Full']

_MAX_WORKERS = 50
_threads_queues = weakref.WeakKeyDictionary()
passthrough = lambda: None


def set_as_loop(loop):
    if loop._iothreadloop:
        asyncio.set_event_loop(loop)


def get_executor(loop):
    executor = loop._default_executor
    if executor is None:
        executor = ThreadPool(loop=loop)
        loop._default_executor = executor
    return executor


def run_in_executor(loop, executor, callback, *args):
    if isinstance(callback, events.Handle):
        assert not args
        assert not isinstance(callback, events.TimerHandle)
        if callback._cancelled:
            f = Future(loop=loop)
            f.set_result(None)
            return f
        callback, args = callback._callback, callback._args
    if executor is None:
        executor = get_executor(loop)
    return executor.submit(callback, *args)


class Thread(dummy.DummyProcess):
    @property
    def pid(self):
        return current_process().pid

    def loop(self):
        raise NotImplemented

    def terminate(self):
        '''Invoke the stop on the event loop method.'''
        if self.is_alive():
            loop = self.loop()
            if loop:
                loop.stop()


class PoolThread(Thread):
    '''A thread for the :class;`.ThreadPool`.
    '''
    def __init__(self, pool):
        self.pool = pool
        super(PoolThread, self).__init__(name=pool.worker_name)

    def __repr__(self):
        if self.ident:
            return '%s-%s' % (self.name, self.ident)
        else:
            return self.name
    __str__ = __repr__

    def run(self):
        '''Modified run method which set the actor and the event_loop.
        '''
        if self.pool._actor:
            set_actor(self.pool._actor)
        asyncio.set_event_loop(self.pool._loop)
        # The run method for the threads in this thread pool
        logger = logging.getLogger('pulsar.%s' % self.name)
        loop = QueueEventLoop(self.pool, logger=logger, iothreadloop=True)
        loop.run_forever()

    def loop(self):
        return thread_data('_request_loop', ct=self)


class IOqueue(selectors.BaseSelector):
    '''A selector based on a distributed queue

    Since there is no way to my knowledge to wake up the queue while
    getting an itiem from the task queue, the timeout cannot be larger than
    a small number which by default is ``0.5`` seconds.
    '''
    max_timeout = 0.5

    def __init__(self, executor):
        super(IOqueue, self).__init__()
        self._actor = executor._actor
        self._work_queue = executor._work_queue
        self._maxtasks = executor._maxtasks
        self._received = 0
        self._completed = 0

    def select(self, timeout=None):
        if self._actor and self._actor.state > ACTOR_STATES.RUN:
            raise _StopError
        if self._maxtasks and self._received >= self._maxtasks:
            if self._completed < self._received:
                return ()
            else:
                raise _StopError
        block = True
        if timeout is None:
            timeout = self.max_timeout
        elif timeout <= 0:
            timeout = 0
            block = False
        else:
            timeout = min(self.max_timeout, timeout)
        try:
            task = self._work_queue.get(block=block, timeout=timeout)
        except (Empty, TypeError):
            return ()
        except (EOFError, IOError):
            raise _StopError
        if task is None:    # got the sentinel, exit!
            self._work_queue.put(None)
            raise _StopError
        return task

    def process_task(self, task):
        self._received += 1
        future, func, args, kwargs = task
        try:
            result = yield func(*args, **kwargs)
        except Exception as exc:
            self._completed += 1
            future.set_exception(exc)
        else:
            self._completed += 1
            try:
                future.set_result(result)
            finally:
                # IMPORTANT: make sure to wake up the loop of the
                # waiting future
                future._loop.call_soon_threadsafe(passthrough)

    def get_map(self):
        return {}

    def register(self, fileobj, events, data=None):
        pass

    def unregister(self, fileobj):
        pass


class QueueEventLoop(BaseEventLoop):
    task_factory = Task

    def __init__(self, executor, iothreadloop=False, logger=None):
        super(QueueEventLoop, self).__init__()
        self._default_executor = executor
        self._iothreadloop = iothreadloop
        self._selector = IOqueue(executor)
        self.logger = logger or LOGGER
        self.call_soon(set_as_loop, self)

    def _write_to_self(self):
        pass

    def _process_events(self, task):
        if task:
            async(self._selector.process_task(task), self)

    def run_in_executor(self, executor, callback, *args):
        return run_in_executor(self, executor, callback, *args)


class ThreadPool(AsyncObject):
    '''A thread pool for an actor.

    This pool maintains a group of threads to perform asynchronous tasks via
    the :meth:`submit` method.
    '''
    worker_name = 'exec'

    def __init__(self, max_workers=None, actor=None, loop=None,
                 maxtasks=None):
        self._actor = actor = actor or get_actor()
        if actor:
            loop = loop or actor._loop
            if not max_workers:
                max_workers = actor.cfg.thread_workers
            self.worker_name = '%s.%s' % (actor.name, self.worker_name)
        self._loop = loop or asyncio.get_event_loop()
        self._max_workers = min(max_workers or _MAX_WORKERS, _MAX_WORKERS)
        self._threads = set()
        self._maxtasks = maxtasks
        self._work_queue = ThreadQueue()
        self._shutdown = False
        self._shutdown_lock = threading.Lock()

    def submit(self, func, *args, **kwargs):
        '''Equivalent to ``func(*args, **kwargs)``.

        This method create a new task for function ``func`` and adds it to
        the queue.
        Return a :class:`~asyncio.Future` called back once the task
        has finished.
        '''
        with self._shutdown_lock:
            if self._shutdown:
                raise RuntimeError(
                    'cannot schedule new futures after shutdown')
            future = Future(loop=self._loop)
            self._work_queue.put((future, func, args, kwargs))
            self._adjust_thread_count()
            return future

    def shutdown(self, wait=True):
        with self._shutdown_lock:
            self._shutdown = True
            self._work_queue.put(None)
        if wait:
            for t in self._threads:
                t.join()

    def _adjust_thread_count(self):
        if len(self._threads) < self._max_workers:
            t = PoolThread(self)
            t.daemon = True
            t.start()
            self._threads.add(t)
            _threads_queues[t] = self._work_queue



'''\
Pulsar Python Shell example::

    python pshell.py
'''
try:
    from pulsar.apps.shell import PulsarShell
except ImportError:  # pragma nocover
    import sys
    sys.path.append('../../')
    from pulsar.apps.shell import PulsarShell


if __name__ == '__main__':  # pragma nocover
    PulsarShell().start()



import sys
import os
from multiprocessing import forking, process, freeze_support
from multiprocessing.util import _logger, _log_to_stderr

WINEXE = forking.WINEXE


def get_preparation_data(name):
    '''
    Return info about parent needed by child to unpickle process object.
    Monkey-patch from
    '''
    d = dict(
        name=name,
        sys_path=sys.path,
        sys_argv=sys.argv,
        log_to_stderr=_log_to_stderr,
        orig_dir=process.ORIGINAL_DIR,
        authkey=process.current_process().authkey,
    )

    if _logger is not None:
        d['log_level'] = _logger.getEffectiveLevel()

    if not WINEXE:
        main_path = getattr(sys.modules['__main__'], '__file__', None)
        if not main_path and sys.argv[0] not in ('', '-c'):
            main_path = sys.argv[0]
        if main_path is not None:
            if (not os.path.isabs(main_path) and process.ORIGINAL_DIR
                    is not None):
                main_path = os.path.join(process.ORIGINAL_DIR, main_path)
            if not main_path.endswith('.exe'):
                d['main_path'] = os.path.normpath(main_path)

    return d


forking.get_preparation_data = get_preparation_data
freeze_support()



'''
A :ref:`task queue <apps-taskqueue>` application implements several
:class:`Job` classes which specify the way a :class:`.Task` is run.
Each :class:`Job` class is a :class:`.Task` factory, therefore,
a :class:`.Task` is always associated
with one :class:`Job`, which can be of two types:

* standard (:class:`.Job`)
* periodic (:class:`.PeriodicJob`), a generator of scheduled tasks.

.. _job-callable:

Job callable method
~~~~~~~~~~~~~~~~~~~~~~~~~~

To define a job is simple, subclass from :class:`.Job` and implement the
**job callable method**::

    from pulsar.apps import tasks

    class Addition(tasks.Job):

        def __call__(self, consumer, a=0, b=0):
            "Add two numbers"
            return a+b

The ``consumer``, instance of :class:`.TaskConsumer`,
is passed by the :ref:`Task backend <apps-taskqueue-backend>` and should
always be the first positional parameter in the callable method.
The remaining (optional key-valued only!) parameters are needed by
your job implementation.

A :ref:`job callable <job-callable>` can also return a
:ref:`coroutine <coroutine>` if it needs to perform asynchronous IO during its
execution::

    class Crawler(tasks.Job):

        def __call__(self, consumer, sample=100, size=10):
            response = yield http.request(...)
            content = response.content
            ...

This allows for cooperative task execution.

.. _job-non-overlap:

Non overlapping Jobs
~~~~~~~~~~~~~~~~~~~~~~~~~~

The :attr:`~.Job.can_overlap` attribute controls the way tasks are generated
by a specific :class:`.Job`. By default, a :class:`.Job` creates a new task
every time the :class:`.TaskBackend` requests it.

However, when setting the :attr:`~.Job.can_overlap` attribute to ``False``,
a new task cannot be started unless a previous task of the same job
is done.

'''
from datetime import datetime, date
import logging

from pulsar.utils.pep import iteritems
from pulsar.utils.importer import import_modules


__all__ = ['JobMetaClass', 'Job', 'PeriodicJob',
           'anchorDate', 'JobRegistry']


class JobRegistry(dict):
    """Site registry for tasks."""

    def regular(self):
        """A generator of all regular jobs."""
        return self.filter_types("regular")

    def periodic(self):
        """A generator of all periodic jobs."""
        return self.filter_types("periodic")

    def register(self, job):
        """Register a job in the job registry.

        The task will be automatically instantiated if not already an
        instance.

        """
        if isinstance(job, JobMetaClass) and job.can_register:
            name = job.name
            self[name] = job()

    def filter_types(self, type):
        """Return a generator of all tasks of a specific type."""
        return ((job_name, job) for job_name, job in iteritems(self)
                if job.type == type)

    @classmethod
    def load(cls, paths):
        self = cls()
        for mod in import_modules(paths):
            for name in dir(mod):
                self.register(getattr(mod, name))
        return self


class JobMetaClass(type):

    def __new__(cls, name, bases, attrs):
        attrs['can_register'] = not attrs.pop('abstract', False)
        job_name = attrs.get("name", name).lower()
        log_prefix = attrs.get("log_prefix") or "pulsar"
        attrs["name"] = job_name
        logname = '%s.job.%s' % (log_prefix, name)
        attrs['logger'] = logging.getLogger(logname)
        return super(JobMetaClass, cls).__new__(cls, name, bases, attrs)


class Job(JobMetaClass('JobBase', (object,), {'abstract': True})):
    '''The Job class which is used in a distributed task queue.

.. attribute:: name

    The unique name which defines the Job and which can be used to retrieve
    it from the job registry. This attribute is set to the Job class name
    in lower case by default, unless a ``name`` class attribute is defined.

.. attribute:: abstract

    If set to ``True`` (default is ``False``), the :class:`.Job` won't be
    registered with the :class:`.JobRegistry`. Useful when creating a new
    base class for several other jobs.

.. attribute:: type

    Type of Job, one of ``regular`` and ``periodic``.

.. attribute:: timeout

    An instance of a datetime.timedelta or ``None``. If set, it represents the
    time lag after which a task which did not start expires.

    Default: ``None``.

.. attribute:: can_overlap

    Boolean indicating if this job can generate overlapping tasks. It can
    also be a callable which accept the same input parameters as the job
    callable function.

    Default: ``True``.

.. attribute:: doc_syntax

    The doc string syntax.

    Default: ``markdown``

.. attribute:: logger

    an instance of a logger. Created at runtime.
'''
    abstract = True
    timeout = None
    expires = None
    doc_syntax = 'markdown'
    can_overlap = True

    def __call__(self, consumer, *args, **kwargs):
        raise NotImplementedError("Jobs must implement the __call__ method.")

    @property
    def type(self):
        '''Type of Job, one of ``regular`` and ``periodic``.'''
        return 'regular'

    def queue_task(self, consumer, jobname, meta_params=None, **kwargs):
        '''Queue a new task in the task queue.

        This utility method can be used from within the
        :ref:`job callable <job-callable>` method and it allows tasks to act
        as tasks factories.

        :parameter consumer: the :class:`.TaskConsumer`
            handling the :class:`.Task`.
            Must be the same instance as the one passed to the
            :ref:`job callable <job-callable>` method.
        :parameter jobname: The name of the :class:`.Job` to run.
        :parameter kwargs: key-valued parameters for the
            :ref:`job callable <job-callable>`.
        :return: a :class:`~asyncio.Future` called back with the task id.

        This method invokes the :meth:`.TaskBackend.queue_task`
        method with the additional ``from_task`` argument equal to the
        id of the task invoking the method.
        '''
        if meta_params is None:
            meta_params = {}
        meta_params['from_task'] = consumer.task_id
        return consumer.backend.queue_task(jobname, meta_params, **kwargs)


class PeriodicJob(Job):
    '''A periodic :class:`.Job` implementation.'''
    abstract = True
    anchor = None
    '''If specified it must be a :class:`~datetime.datetime` instance.
    It controls when the periodic Job is run.
    '''
    run_every = None
    '''Periodicity as a :class:`~datetime.timedelta` instance.'''

    def __init__(self, run_every=None):
        self.run_every = run_every or self.run_every
        if self.run_every is None:
            raise NotImplementedError('Periodic Jobs must have a run_every '
                                      'attribute set, "{0}" does not have one'
                                      .format(self.name))

    @property
    def type(self):
        return 'periodic'

    def is_due(self, last_run_at):
        """Returns tuple of two items ``(is_due, next_time_to_run)``,
        where next time to run is in seconds. For example:

        * ``(True, 20)``, means the job should be run now, and the next
          time to run is in 20 seconds.

        * ``(False, 12)``, means the job should be run in 12 seconds.

        You can override this to decide the interval at runtime.
        """
        return self.run_every.is_due(last_run_at)


def anchorDate(hour=0, minute=0, second=0):
    '''Create an anchor date.'''
    td = date.today()
    return datetime(year=td.year, month=td.month, day=td.day,
                    hour=hour, minute=minute, second=second)



'''
.. _wsgi-response-middleware:

Response middleware are callable objects which can be used in conjunction
with pulsar :ref:`WsgiHandler <wsgi-handler>`.
They must return a :ref:`WsgiResponse <wsgi-response>` which can be the
same as the one passed to the callable or a brand new one.

Interface
=================

.. autoclass:: ResponseMiddleware
   :members:
   :member-order: bysource

GZip Middleware
=================
.. autoclass:: GZipMiddleware
   :members:
   :member-order: bysource


AccessControl
=================
.. autoclass:: AccessControl
   :members:
   :member-order: bysource

'''
import re
from gzip import GzipFile

from pulsar.utils.httpurl import BytesIO


re_accepts_gzip = re.compile(r'\bgzip\b')


__all__ = ['AccessControl', 'GZipMiddleware']


class ResponseMiddleware(object):
    '''Base class for response middlewares.

    A response middleware is used by a :ref:`WsgiHandler <wsgi-handler>`,
    it is a callable used to manipulate a :ref:`WsgiResponse <wsgi-response>`.

    The focus of this class is the :meth:`execute` method where
    the middleware logic is implemented.
    '''
    def version(self, environ):
        return environ.get('wsgi.version')

    def available(self, environ, response):
        '''Check if this :class:`ResponseMiddleware` can be applied to
        the ``response`` object.

        :param environ: a WSGI environ dictionary.
        :param response: a :class:`pulsar.apps.wsgi.wrappers.WsgiResponse`
        :return: ``True`` or ``False``.
        '''
        return True

    def __call__(self, environ, response):
        if not self.available(environ, response):
            return response
        resp = self.execute(environ, response)
        return resp if resp is not None else response

    def execute(self, environ, response):
        '''Manipulate *response*, called only if the :meth:`available`
method returns ``True``.'''
        pass


class AccessControl(ResponseMiddleware):
    '''A response middleware which add the ``Access-Control-Allow-Origin``
    response header.
    '''
    def __init__(self, origin='*', methods=None):
        self.origin = origin
        self.methods = methods

    def available(self, environ, response):
        return response.status_code == 200

    def execute(self, environ, response):
        response.headers['Access-Control-Allow-Origin'] = self.origin
        if self.methods:
            response.headers['Access-Control-Allow-Methods'] = self.methods


class GZipMiddleware(ResponseMiddleware):
    """A :class:`ResponseMiddleware` for compressing content if the request
allows gzip compression. It sets the Vary header accordingly.

The compression implementation is from
http://jython.xhaus.com/http-compression-in-python-and-jython
    """
    def __init__(self, min_length=200):
        self.min_length = min_length

    def available(self, environ, response):
        # It's not worth compressing non-OK or really short responses
        try:
            if response.status_code == 200 and not response.is_streamed:
                if response.length() < self.min_length:
                    return False
                headers = response.headers
                # Avoid gzipping if we've already got a content-encoding.
                if 'Content-Encoding' in headers:
                    return False
                # MSIE have issues with gzipped response of various
                # content types.
                if "msie" in environ.get('HTTP_USER_AGENT', '').lower():
                    ctype = headers.get('Content-Type', '').lower()
                    if not ctype.startswith("text/") or "javascript" in ctype:
                        return False
                ae = environ.get('HTTP_ACCEPT_ENCODING', '')
                if not re_accepts_gzip.search(ae):
                    return False
                return True
        except Exception:
            raise

    def execute(self, environ, response):
        headers = response.headers
        headers.add_header('Vary', 'Accept-Encoding')
        content = b''.join(response.content)
        response.content = (self.compress_string(content),)
        response.headers['Content-Encoding'] = 'gzip'

    def compress_string(self, s):
        zbuf = BytesIO()
        zfile = GzipFile(mode='wb', compresslevel=6, fileobj=zbuf)
        zfile.write(s)
        zfile.close()
        return zbuf.getvalue()



'''
Importing this module to add this extra settings to pulsar
'''
import pulsar


class BackendServer(pulsar.Setting):
    name = 'backend_server'
    flags = ['-s', '--backend-server']
    meta = "CONNECTION STRING"
    default = ''
    validator = pulsar.validate_string
    desc = 'Connection string to a backend server'



'''
This module implements the main classes for pulsar application framework. The
framework is built on top of pulsar asynchronous engine and allows to
implement servers with very little effort. The main classes here are:
:class:`Application` and :class:`MultiApp` which, has the name suggests, is
a factory of several :class:`Application` running on a single server.
The :class:`Configurator` is a mixin used as base class for both
:class:`Application` and :class:`MultiApp`.


Configurator
===============================

.. autoclass:: Configurator
   :members:
   :member-order: bysource


Application
===============================

.. autoclass:: Application
   :members:
   :member-order: bysource

   .. automethod:: __call__


Multi App
===============================

.. autoclass:: MultiApp
   :members:
   :member-order: bysource


Get application
=========================

.. autofunction:: get_application


.. _when-monitor-start:

When monitor start
=================================

The application framework provides a way for adding hooks which are executed
every time a new application starts. A hook is registered by::

    from pulsar.apps import when_monitor_start

    def myhook(monitor):
        ...

    when_monitor_start.append(myhook)

By default, the list of hooks only contains a callback to start the
:ref:`default data store <setting-data_store>` if it needs to.
'''
import os
import sys
from hashlib import sha1
from inspect import getfile
from functools import partial
from collections import namedtuple

import pulsar
from pulsar import (get_actor, coroutine_return, Config, async,
                    multi_async, Future, ImproperlyConfigured)
from pulsar.utils.structures import OrderedDict

__all__ = ['Application', 'MultiApp', 'get_application', 'when_monitor_start']


when_monitor_start = []
new_app = namedtuple('new_app', 'prefix params')


def get_application(name):
    '''Fetch an :class:`Application` associated with ``name`` if available.

    This function may return an :ref:`asynchronous component <coroutine>`.
    The application name is set during initialisation. Check the
    :attr:`Configurator.name` attribute for more information.
    '''
    actor = get_actor()

    if actor:
        if actor.is_arbiter():
            return _get_app(actor, name, False)
        else:
            return _get_remote_app(actor, name)


def _get_remote_app(actor, name):
    cfg = yield actor.send('arbiter', 'run', _get_app, name)
    coroutine_return(cfg.app() if cfg else None)


def _get_app(arbiter, name, safe=True):
    monitor = arbiter.get_actor(name)
    if monitor:
        cfg = yield monitor.start_event
        if safe:
            coroutine_return(cfg)
        else:
            coroutine_return(monitor.app)


def monitor_start(self, exc=None):
    start_event = self.start_event
    if exc:
        start_event.set_exception(exc)
        return
    app = self.app
    try:
        self.bind_event('on_params', monitor_params)
        self.bind_event('on_info', monitor_info)
        self.bind_event('stopping', monitor_stopping)
        for callback in when_monitor_start:
            yield callback(self)
        self.monitor_task = lambda: app.monitor_task(self)
        yield app.monitor_start(self)
        if not self.cfg.workers:
            yield app.worker_start(self)
        result = self.cfg
    except Exception as exc:
        yield self.stop(exc)
        start_event.set_result(None)
    else:
        start_event.set_result(result)


def monitor_stopping(self, exc=None):
    if not self.cfg.workers:
        yield self.app.worker_stopping(self)
    yield self.app.monitor_stopping(self)
    coroutine_return(self)


def monitor_info(self, info=None):
    if not self.cfg.workers:
        self.app.worker_info(self, info)
    else:
        self.app.monitor_info(self, info)


def monitor_params(self, params=None):
    app = self.app
    params.update({'cfg': app.cfg.clone(),
                   'name': '%s.worker' % app.name,
                   'start': worker_start})
    app.actorparams(self, params)


def worker_start(self, exc=None):
    app = getattr(self, 'app', None)
    if app is None:
        cfg = self.cfg
        self.app = app = cfg.application.from_config(cfg, logger=self.logger)
    self.bind_event('on_info', app.worker_info)
    self.bind_event('stopping', app.worker_stopping)
    return app.worker_start(self, exc=exc)


class Configurator(object):
    '''A mixin for configuring and loading a pulsar application server.

    :parameter name: to override the class :attr:`name` attribute.
    :parameter description: to override the class :attr:`cfg.description`
        attribute.
    :parameter epilog: to override the class :attr:`cfg.epilog` attribute.
    :parameter version: Optional version of this application, it overrides
        the class :attr:`cfg.version` attribute.
    :parameter argv: Optional list of command line parameters to parse, if
        not supplied the :attr:`sys.argv` list will be used. The parameter
        is only relevant if ``parse_console`` is ``True``.
    :parameter parse_console: ``True`` (default) if the console parameters
        needs parsing.
    :parameter script: Optional string which set the :attr:`script`
        attribute.
    :parameter params: a dictionary of configuration parameters which
        overrides the defaults and the :attr:`cfg` class attribute.
        They will be overwritten by a :ref:`config file <setting-config>`
        or command line arguments.

    .. attribute:: name

        The name is unique if this is an :class:`Application`. In this
        case it defines the application monitor name as well and can be
        access in the arbiter domain via the :func:`get_application`
        function.

    .. attribute:: argv

        Optional list of command line parameters. If not available the
        :attr:`sys.argv` list will be used when parsing the console.

    .. attribute:: cfg

        The :class:`.Config` for this :class:`Configurator`.
        If set as class attribute it will be replaced during initialisation.

        Default: ``None``.

    .. attribute:: console_parsed

        ``True`` if this application parsed the console before starting.

    .. attribute:: script

        Full path of the script which starts the application or ``None``.
        Evaluated during initialization via the :meth:`python_path` method.
    '''
    argv = None
    name = None
    cfg = Config()

    def __init__(self,
                 name=None,
                 description=None,
                 epilog=None,
                 version=None,
                 argv=None,
                 parse_console=True,
                 script=None,
                 cfg=None,
                 load_config=True,
                 **params):
        cls = self.__class__
        self.name = name or cls.name or cls.__name__.lower()
        if not isinstance(cfg, Config):
            cfg = cfg or {}
            cfg.update(params)
            cfg = cls.create_config(cfg)
        else:
            cfg.update(params)
        self.cfg = cfg
        cfg.description = description or self.cfg.description
        cfg.epilog = epilog or self.cfg.epilog
        cfg.version = version or self.cfg.version
        cfg.name = self.name
        cfg.application = cls
        self.argv = argv
        self.console_parsed = parse_console
        self.cfg.script = self.script = self.python_path(script)

    @property
    def version(self):
        '''Version of this :class:`Application`'''
        return self.cfg.version

    @property
    def root_dir(self):
        '''Root directory of this :class:`Configurator`.

        Evaluated from the :attr:`script` attribute.
        '''
        if self.cfg.script:
            return os.path.dirname(self.cfg.script)

    def __repr__(self):
        return self.name

    def __str__(self):
        return self.__repr__()

    def python_path(self, script):
        '''Called during initialisation to obtain the ``script`` name and
        to add the :attr:`script` directory to the python path if not in the
        path already.
        If ``script`` does not evalueate to ``True`` it is evaluated from
        the ``__main__`` import. Returns the real path of the python
        script which runs the application.
        '''
        if not script:
            try:
                import __main__
                script = getfile(__main__)
            except Exception:  # pragma    nocover
                return
        script = os.path.realpath(script)
        path = os.path.dirname(script)
        if path not in sys.path:
            sys.path.insert(0, path)
        return script

    def on_config(self, arbiter):
        '''Callback when configuration is loaded.

        This is a chance to do applications specific checks before the
        concurrent machinery is put into place.

        If it returns ``False`` the application will abort.
        '''
        pass

    def load_config(self):
        '''Load the application configuration from a file and/or
        from the command line.

        Called during application initialisation. The parameters
        overriding order is the following:

        * default parameters.
        * the key-valued params passed in the initialisation.
        * the parameters in the optional configuration file
        * the parameters passed in the command line.
        '''
        # get the actor if available and override default cfg values with those
        # from the actor
        actor = get_actor()
        cfg = self.cfg
        if actor and actor.is_running():
            # actor available and running.
            # Unless argv is set, skip parsing
            if self.argv is None:
                self.console_parsed = False
            # copy global settings
            self.cfg.copy_globals(actor.cfg)
        #
        for name in list(self.cfg.params):
            if name in self.cfg.settings:
                value = self.cfg.params.pop(name)
                if value is not None:
                    self.cfg.set(name, value)
        # parse console args
        if self.console_parsed:
            parser = self.cfg.parser()
            opts = parser.parse_args(self.argv)
            config = getattr(opts, 'config', None)
            # set the config only if config is part of the settings
            if config is not None and self.cfg.config:
                self.cfg.set('config', config)
        else:
            parser, opts = None, None
        #
        # Load up the config file if found.
        self.cfg.params.update(self.cfg.import_from_module())
        #
        # Update the configuration with any command line settings.
        if opts:
            for k, v in opts.__dict__.items():
                if v is None:
                    continue
                self.cfg.set(k.lower(), v)

    def start(self):
        '''Invoked the application callable method and start
        the :class:`.Arbiter` if it wasn't already started.

        It returns a :class:`~asyncio.Future` called back once the
        application/applications are running. It returns ``None`` if
        called more than once.
        '''
        on_start = self()
        arbiter = pulsar.arbiter()
        if arbiter and on_start:
            arbiter.start()
        return on_start

    @classmethod
    def create_config(cls, params, prefix=None, name=None):
        '''Create a new :class:`.Config` container.

        Invoked during initialisation, it overrides defaults
        with ``params`` and apply the ``prefix`` to non global
        settings.
        '''
        if isinstance(cls.cfg, Config):
            cfg = cls.cfg.copy(name=name, prefix=prefix)
        else:
            cfg = cls.cfg.copy()
            if name:
                cfg[name] = name
            if prefix:
                cfg[prefix] = prefix
            cfg = Config(**cfg)
        cfg.update_settings()
        cfg.update(params)
        return cfg


class Application(Configurator):
    """An application interface.

    Applications can be of any sorts or forms and the library is shipped with
    several battery included examples in the :mod:`pulsar.apps` module.

    These are the most important facts about a pulsar :class:`Application`:

    * It derives from :class:`Configurator` so that it has all the
      functionalities to parse command line arguments and setup the
      :attr:`~Configurator.cfg` placeholder of :class:`.Setting`.
    * Instances of an :class:`Application` are callable objects accepting
      the calling actor as only argument. The callable method should
      not be overwritten, instead one should overwrites the application
      hooks available.
    * When an :class:`Application` is called for the first time,
      a new :class:`.Monitor` instance is added to the
      :class:`.Arbiter`, ready to perform its duties.

    :parameter callable: Initialise the :attr:`callable` attribute.
    :parameter load_config: If ``False`` the :meth:`~Configurator.load_config`
        method is not invoked.

        Default ``True``.
    :parameter params: Passed to the :class:`Configurator` initialiser.

    .. attribute:: callable

        Optional callable serving or configuring your application.
        If provided, the callable must be picklable, therefore it is either
        a function or a picklable object.

        Default ``None``
    """
    def __init__(self, callable=None, load_config=True, **params):
        super(Application, self).__init__(load_config=load_config, **params)
        self.cfg.callable = callable
        if load_config:
            self.load_config()

    @classmethod
    def from_config(cls, cfg, logger=None):
        c = cls.__new__(cls)
        c.name = cfg.name
        c.cfg = cfg
        c.logger = logger or cfg.configured_logger()
        return c

    @property
    def stream(self):
        '''Actor stream handler.
        '''
        return get_actor().stream

    def __call__(self, actor=None):
        '''Register this application with the (optional) calling ``actor``.

        If an ``actor`` is available (either via the function argument or via
        the :func:`~pulsar.async.actor.get_actor` function) it must be
        :class:`.Arbiter`, otherwise this call is no-op.

        If no actor is available, it means this application starts
        pulsar engine by creating the :class:`.Arbiter` with its
        :ref:`global settings <setting-section-global-server-settings>`
        copied to the arbiter :class:`.Config` container.

        :return: the ``start`` one time event fired once this application
            has fired it.
        '''
        if actor is None:
            actor = get_actor()
        monitor = None
        if actor and actor.is_arbiter():
            monitor = actor.get_actor(self.name)
        if monitor is None and (not actor or actor.is_arbiter()):
            self.cfg.on_start()
            self.logger = self.cfg.configured_logger()
            if not actor:
                actor = pulsar.arbiter(cfg=self.cfg.clone())
            else:
                self.update_arbiter_params(actor)
            if not self.cfg.exc_id:
                self.cfg.set('exc_id', actor.cfg.exc_id)
            if self.on_config(actor) is not False:
                start = Future(loop=actor._loop)
                actor.bind_event('start', partial(self._add_monitor, start))
                return start
            else:
                return
        raise ImproperlyConfigured('Already started or not in arbiter domain')

    # WORKERS CALLBACKS
    def worker_start(self, worker, exc=None):
        '''Added to the ``start`` :ref:`worker hook <actor-hooks>`.'''
        pass

    def worker_info(self, worker, info):
        '''Hook to add additional entries to the worker ``info`` dictionary.
        '''
        pass

    def worker_stopping(self, worker, exc=None):
        '''Added to the ``stopping`` :ref:`worker hook <actor-hooks>`.'''
        pass

    # MONITOR CALLBACKS
    def actorparams(self, monitor, params=None):
        '''Hook to add additional entries when the monitor spawn new actors.
        '''
        pass

    def monitor_start(self, monitor):
        '''Callback by the monitor when starting.
        '''
        pass

    def monitor_info(self, monitor, info):
        '''Hook to add additional entries to the monitor ``info`` dictionary.
        '''
        pass

    def monitor_stopping(self, monitor):
        '''Callback by the monitor before stopping.
        '''
        pass

    def monitor_task(self, monitor):
        '''Executed by the :class:`.Monitor` serving this application
        at each event loop.'''
        pass

    def update_arbiter_params(self, arbiter):
        for s in self.cfg.settings.values():
            if s.is_global and s.modified:
                a = arbiter.cfg.settings[s.name]
                if not a.modified:
                    a.set(s.value)

    #   INTERNALS
    def _add_monitor(self, start, arbiter, exc=None):
        if not exc:
            monitor = arbiter.add_monitor(
                self.name, app=self, cfg=self.cfg,
                start=monitor_start, start_event=start)
            self.cfg = monitor.cfg


class MultiApp(Configurator):
    '''A :class:`MultiApp` is a tool for creating several :class:`Application`
    and starting them at once.

    It makes sure all :ref:`settings <settings>` for the
    applications created are available in the command line.
    Check the :class:`~examples.taskqueue.manage.server` class in the
    :ref:`taskqueue example <tutorials-taskqueue>` for an actual
    implementation.

    :class:`MultiApp` derives from :class:`Configurator` and therefore
    supports all its configuration utilities,
    :meth:`build` is the only method which must be implemented by
    subclasses.

    A minimal example usage::

        import pulsar

        class Server(pulsar.MultiApp):

            def build(self):
                yield self.new_app(TaskQueue)
                yield self.new_app(WSGIserver, prefix="rpc", callable=..., ...)
                yield self.new_app(WSGIserver, prefix="web", callable=..., ...)
    '''
    _apps = None

    def build(self):
        '''Virtual method, must be implemented by subclasses and return an
        iterable over results obtained from calls to the
        :meth:`new_app` method.
        '''
        raise NotImplementedError

    def apps(self):
        '''List of :class:`Application` for this :class:`MultiApp`.

        The list is lazily loaded from the :meth:`build` method.
        '''
        if self._apps is None:
            # Add modified settings values to the list of cfg params
            self.cfg.params.update(((s.name, s.value) for s in
                                    self.cfg.settings.values() if s.modified))
            self.cfg.settings = {}
            self._apps = OrderedDict()
            self._apps.update(self._build())
            if not self._apps:
                return []
            # Load the configuration (command line and config file)
            self.load_config()
            kwargs = self._get_app_params()
            apps = self._apps
            self._apps = []
            for App, name, callable, cfg in self._iter_app(apps):
                settings = self.cfg.settings
                new_settings = {}
                for key in cfg:
                    setting = settings[key].copy()
                    if setting.orig_name and setting.orig_name != setting.name:
                        setting.name = setting.orig_name
                    new_settings[setting.name] = setting
                cfg.settings = new_settings
                kwargs.update({'name': name, 'cfg': cfg, 'callable': callable})
                if name == self.name:
                    params = kwargs.copy()
                    params['version'] = self.version
                else:
                    params = kwargs
                self._apps.append(App(**params))
        return self._apps

    def new_app(self, App, prefix=None, callable=None, **params):
        '''Invoke this method in the :meth:`build` method as many times
        as the number of :class:`Application` required by this
        :class:`MultiApp`.

        :param App: an :class:`Application` class.
        :param prefix: The prefix to use for the application,
            the prefix is appended to
            the application :ref:`config parameters <settings>` and to the
            application name. Each call to this methjod must use a different
            value of for this parameter. It can be ``None``.
        :param callable: optional callable (function of object) used during
            initialisation of *App* (the :class:`Application.callable`).
        :param params: additional key-valued parameters used when creating
            an instance of *App*.
        :return: a tuple used by the :meth:`apps` method.
        '''
        params.update(self.cfg.params.copy())
        params.pop('name', None)    # remove the name
        prefix = prefix or ''
        if not prefix and '' in self._apps:
            prefix = App.name or App.__name__.lower()
        if not prefix:
            name = self.name
            cfg = App.create_config(params, name=name)
        else:
            name = '%s_%s' % (prefix, self.name)
            cfg = App.create_config(params, prefix=prefix, name=name)
        # Add the config entry to the multi app config if not available
        for k in cfg.settings:
            if k not in self.cfg.settings:
                self.cfg.settings[k] = cfg.settings[k]
        return new_app(prefix, (App, name, callable, cfg))

    def __call__(self, actor=None):
        apps = [app(actor) for app in self.apps()]
        return multi_async(apps, loop=get_actor()._loop)

    #    INTERNALS
    def _build(self):
        for app in self.build():
            if not isinstance(app, new_app):
                raise ImproperlyConfigured(
                    'You must use new_app when building a MultiApp')
            yield app

    def _iter_app(self, app_name_callables):
        main = app_name_callables.pop('', None)
        if not main:
            raise ImproperlyConfigured('No main application in MultiApp')
        yield main
        for app in app_name_callables.values():
            yield app

    def _get_app_params(self):
        params = self.cfg.params.copy()
        for key, value in self.__dict__.items():
            if key.startswith('_'):
                continue
            elif key == 'console_parsed':
                params['parse_console'] = not value
            else:
                params[key] = value
        params['load_config'] = False
        return params



import pulsar
from pulsar import coroutine_return

from .jsonrpc import JSONRPC


__all__ = ['PulsarServerCommands']


class PulsarServerCommands(JSONRPC):
    '''Useful commands to add to your :class:`.JSONRPC` handler.

    It exposes the following functions:'''
    def rpc_ping(self, request):
        '''Ping the server.'''
        return 'pong'

    def rpc_echo(self, request, message=''):
        '''Echo the server.'''
        return message

    def rpc_server_info(self, request):
        '''Return a dictionary of information regarding the server and workers.

        It invokes the :meth:`extra_server_info` for adding custom
        information.
        '''
        info = yield pulsar.send('arbiter', 'info')
        info = yield self.extra_server_info(request, info)
        coroutine_return(info)

    def rpc_functions_list(self, request):
        '''List of (method name, method document) pair of all method exposed
        by this :class:`.JSONRPC` handler.'''
        return list(self.listFunctions())

    def rpc_documentation(self, request):
        '''Documentation in restructured text.'''
        return self.docs()

    def rpc_kill_actor(self, request, aid):
        '''Kill the actor with id equal to *aid*.'''
        return pulsar.send('arbiter', 'kill_actor', aid)

    def extra_server_info(self, request, info):
        '''An internal method.

        Used by the :meth:`rpc_server_info` method to add additional
        information to the info dictionary.
        '''
        return info



'''
Parser for multipart/form-data
==============================

This module provides a parser for the multipart/form-data format. It can read
from a file, a socket or a WSGI environment.
'''
import re
import sys
from tempfile import TemporaryFile
from wsgiref.headers import Headers
from base64 import b64encode
from io import BytesIO

from .httpurl import parse_qs, ENCODE_BODY_METHODS, mapping_iterator
from .structures import MultiValueDict


def copy_file(stream, target, maxread=-1, buffer_size=2*16):
    ''' Read from :stream and write to :target until :maxread or EOF. '''
    size, read = 0, stream.read
    while 1:
        to_read = buffer_size if maxread < 0 else min(buffer_size,
                                                      maxread-size)
        part = read(to_read)
        if not part:
            return size
        target.write(part)
        size += len(part)

# ############################################################################
# ############################## Header Parser ###############################
# ############################################################################

_special = re.escape('()<>@,;:\\"/[]?={} \t')
_re_special = re.compile('[%s]' % _special)
_qstr = '"(?:\\\\.|[^"])*"'  # Quoted string
_value = '(?:[^%s]+|%s)' % (_special, _qstr)  # Save or quoted string
_option = '(?:;|^)\s*([^%s]+)\s*=\s*(%s)' % (_special, _value)
_re_option = re.compile(_option)  # key=value part of an Content-Type header


def header_quote(val):
    if not _re_special.search(val):
        return val
    return '"' + val.replace('\\', '\\\\').replace('"', '\\"') + '"'


def header_unquote(val, filename=False):
    if val[0] == val[-1] == '"':
        val = val[1:-1]
        if val[1:3] == ':\\' or val[:2] == '\\\\':
            val = val.split('\\')[-1]  # fix ie6 bug: full path --> filename
        return val.replace('\\\\', '\\').replace('\\"', '"')
    return val


def parse_options_header(header, options=None):
    if ';' not in header:
        return header.lower().strip(), {}
    ctype, tail = header.split(';', 1)
    options = options or {}
    for match in _re_option.finditer(tail):
        key = match.group(1).lower()
        value = header_unquote(match.group(2), key == 'filename')
        options[key] = value
    return ctype, options

# ############################################################################
# ################################ Multipart #################################
# ############################################################################


class MultipartError(ValueError):
    pass


class MultipartParser(object):

    def __init__(self, stream, boundary, content_length=-1,
                 disk_limit=2**30, mem_limit=2**20, memfile_limit=2**18,
                 buffer_size=2**16, charset='latin1'):
        ''' Parse a multipart/form-data byte stream. This object is an
        iterator over the parts of the message.

        :param stream: A file-like stream. Must implement ``.read(size)``.
        :param boundary: The multipart boundary as a byte string.
        :param content_length: The maximum number of bytes to read.
        '''
        self.stream, self.boundary = stream, boundary
        self.content_length = content_length
        self.disk_limit = disk_limit
        self.memfile_limit = memfile_limit
        self.mem_limit = min(mem_limit, self.disk_limit)
        self.buffer_size = min(buffer_size, self.mem_limit)
        self.charset = charset
        if self.buffer_size - 6 < len(boundary):  # "--boundary--\r\n"
            raise MultipartError('Boundary does not fit into buffer_size.')
        self._done = []
        self._part_iter = None
        self.separator = '--{0}'.format(self.boundary).encode()
        self.terminator = '--{0}--'.format(self.boundary).encode()

    def __iter__(self):
        ''' Iterate over the parts of the multipart message. '''
        if not self._part_iter:
            self._part_iter = self._iterparse()
        for part in self._done:
            yield part
        for part in self._part_iter:
            self._done.append(part)
            yield part

    def parts(self):
        ''' Returns a list with all parts of the multipart message. '''
        return list(iter(self))

    def get(self, name, default=None):
        ''' Return the first part with that name or a default value (None). '''
        for part in self:
            if name == part.name:
                return part
        return default

    def get_all(self, name):
        ''' Return a list of parts with that name. '''
        return [p for p in self if p.name == name]

    def _lineiter(self):
        ''' Iterate over a binary file-like object line by line. Each line is
            returned as a (line, line_ending) tuple. If the line does not fit
            into self.buffer_size, line_ending is empty and the rest of the
            lineis returned with the next iteration.
        '''
        read = self.stream.read
        maxread, maxbuf = self.content_length, self.buffer_size
        _bcrnl = b'\r\n'
        _bcr = _bcrnl[:1]
        _bnl = _bcrnl[1:]
        _bempty = _bcrnl[:0]  # b'rn'[:0] -> b''
        buffer = _bempty  # buffer for the last (partial) line
        while 1:
            data = read(maxbuf if maxread < 0 else min(maxbuf, maxread))
            maxread -= len(data)
            lines = (buffer+data).splitlines(True)
            len_first_line = len(lines[0])
            # be sure that the first line does not become too big
            if len_first_line > self.buffer_size:
                # at the same time don't split a '\r\n' accidentally
                if (len_first_line == self.buffer_size+1 and
                        lines[0].endswith(_bcrnl)):
                    splitpos = self.buffer_size - 1
                else:
                    splitpos = self.buffer_size
                lines[:1] = [lines[0][:splitpos],
                             lines[0][splitpos:]]
            if data:
                buffer = lines[-1]
                lines = lines[:-1]
            for line in lines:
                if line.endswith(_bcrnl):
                    yield line[:-2], _bcrnl
                elif line.endswith(_bnl):
                    yield line[:-1], _bnl
                elif line.endswith(_bcr):
                    yield line[:-1], _bcr
                else:
                    yield line, _bempty
            if not data:
                break

    def _iterparse(self):
        lines, line = self._lineiter(), ''
        separator = self.separator
        terminator = self.terminator
        # Consume first boundary. Ignore leading blank lines
        for line, nl in lines:
            if line:
                break
        if line != separator:
            raise MultipartError("Stream does not start with boundary")
        # For each part in stream...
        mem_used, disk_used = 0, 0  # Track used resources to prevent DoS
        is_tail = False  # True if the last line was incomplete (cutted)
        opts = {'buffer_size': self.buffer_size,
                'memfile_limit': self.memfile_limit,
                'charset': self.charset}
        part = MultipartPart(**opts)
        for line, nl in lines:
            if line == terminator and not is_tail:
                part.file.seek(0)
                yield part
                break
            elif line == separator and not is_tail:
                if part.is_buffered():
                    mem_used += part.size
                else:
                    disk_used += part.size
                part.file.seek(0)
                yield part
                part = MultipartPart(**opts)
            else:
                is_tail = not nl  # The next line continues this one
                part.feed(line, nl)
                if part.is_buffered():
                    if part.size + mem_used > self.mem_limit:
                        raise MultipartError("Memory limit reached.")
                elif part.size + disk_used > self.disk_limit:
                    raise MultipartError("Disk limit reached.")
        if line != terminator:
            raise MultipartError("Unexpected end of multipart stream.")


class MultipartPart(object):
    default_charset = 'latin1'

    def __init__(self, buffer_size=2**16, memfile_limit=2**18, charset=None):
        self.headerlist = []
        self.headers = None
        self.file = False
        self.size = 0
        self._buf = b''
        self.disposition, self.name, self.filename = None, None, None
        self.content_type = None
        self.charset = charset or self.default_charset
        self.memfile_limit = memfile_limit
        self.buffer_size = buffer_size

    def feed(self, line, nl=''):
        if self.file:
            return self.write_body(line, nl)
        return self.write_header(line, nl)

    def write_header(self, line, nl):
        line = line.decode(self.charset)
        if not nl:
            raise MultipartError('Unexpected end of line in header.')
        if not line.strip():  # blank line -> end of header segment
            self.finish_header()
        elif line[0] in ' \t' and self.headerlist:
            name, value = self.headerlist.pop()
            self.headerlist.append((name, value+line.strip()))
        else:
            if ':' not in line:
                raise MultipartError("Syntax error in header: No colon.")
            name, value = line.split(':', 1)
            self.headerlist.append((name.strip(), value.strip()))

    def write_body(self, line, nl):
        if not line and not nl:  # This does not even flush the buffer
            return
        self.size += len(line) + len(self._buf)
        self.file.write(self._buf + line)
        self._buf = nl
        if self.content_length > 0 and self.size > self.content_length:
            raise MultipartError('Size of body exceeds Content-Length header.')
        if self.size > self.memfile_limit and isinstance(self.file, BytesIO):
            # TODO: What about non-file uploads that exceed the memfile_limit?
            self.file, old = TemporaryFile(mode='w+b'), self.file
            old.seek(0)
            copy_file(old, self.file, self.size, self.buffer_size)

    def finish_header(self):
        self.file = BytesIO()
        self.headers = Headers(self.headerlist)
        cdis = self.headers.get('Content-Disposition', '')
        ctype = self.headers.get('Content-Type', '')
        clen = self.headers.get('Content-Length', '-1')
        if not cdis:
            raise MultipartError('Content-Disposition header is missing.')
        self.disposition, self.options = parse_options_header(cdis)
        self.name = self.options.get('name')
        self.filename = self.options.get('filename')
        self.content_type, options = parse_options_header(ctype)
        self.charset = options.get('charset') or self.charset
        self.content_length = int(self.headers.get('Content-Length', '-1'))

    def is_buffered(self):
        ''' Return true if the data is fully buffered in memory.'''
        return isinstance(self.file, BytesIO)

    def bytes(self):
        pos = self.file.tell()
        self.file.seek(0)
        val = self.file.read()
        self.file.seek(pos)
        return val

    def base64(self, charset=None):
        '''Data encoded as base 64'''
        return b64encode(self.bytes()).decode(charset or self.charset)

    def string(self, charset=None):
        '''Data decoded with the specified charset'''
        return self.bytes().decode(charset or self.charset)

    def save_as(self, path):
        fp = open(path, 'wb')
        pos = self.file.tell()
        try:
            self.file.seek(0)
            size = copy_file(self.file, fp)
        finally:
            self.file.seek(pos)
        return size


def parse_form_data(environ, charset='utf-8', strict=False, **kw):
    '''Parse form data from an environ dict and return a (forms, files) tuple.
Both tuple values are dictionaries with the form-field name as a key
(unicode) and lists as values (multiple values per key are possible).
The forms-dictionary contains form-field values as unicode strings.
The files-dictionary contains :class:`MultipartPart` instances, either
because the form-field was a file-upload or the value is to big to fit
into memory limits.

:parameter environ: A WSGI environment dict.
:parameter charset: The charset to use if unsure. (default: utf8)
:parameter strict: If True, raise :exc:`MultipartError` on any parsing
    errors. These are silently ignored by default.'''
    forms, files = MultiValueDict(), MultiValueDict()
    try:
        if (environ.get('REQUEST_METHOD', 'GET').upper()
                not in ENCODE_BODY_METHODS):
            raise MultipartError("Request method not valid.")
        content_length = int(environ.get('CONTENT_LENGTH', '-1'))
        content_type = environ.get('CONTENT_TYPE', '')
        if not content_type:
            raise MultipartError("Missing Content-Type header.")
        content_type, options = parse_options_header(content_type)
        stream = environ.get('wsgi.input') or BytesIO()
        kw['charset'] = charset = options.get('charset', charset)
        if content_type == 'multipart/form-data':
            boundary = options.get('boundary', '')
            if not boundary:
                raise MultipartError("No boundary for multipart/form-data.")
            for part in MultipartParser(stream, boundary,
                                        content_length, **kw):
                if part.filename or not part.is_buffered():
                    files[part.name] = part
                else:
                    forms[part.name] = part.string()
        elif content_type in ('application/x-www-form-urlencoded',
                              'application/x-url-encoded'):
            mem_limit = kw.get('mem_limit', 2**20)
            if content_length > mem_limit:
                raise MultipartError("Request to big. Increase MAXMEM.")
            data = stream.read(mem_limit).decode(charset)
            if stream.read(1):  # These is more that does not fit mem_limit
                raise MultipartError("Request to big. Increase MAXMEM.")
            data = parse_qs(data, keep_blank_values=True)
            for key, values in mapping_iterator(data):
                for value in values:
                    forms[key] = value
        else:
            raise MultipartError("Unsupported content type.")
    except MultipartError:
        if strict:
            raise
    return forms, files



"""
Pulsar :mod:`~pulsar.apps.greenio` application can be used in conjunction
with :psycopg2coroutine:`psycopg2 coroutine support <>` to
query the :postgresql:`PostgreSql <>` database in an implicit
asynchronous fashion.
Such feature can be extremely useful since it allows to write
asynchronous code without the need to explicitly yield
control appropriately when necessary.

For example, one could use ORMs (Object Relational Mappers) such as
:sqlalchemy:`SqlAlchemy <>` and :django:`django <>` in asynchronous mode
by invoking :func:`make_asynchronous` somehere convinient in the code::

    from pulsar.apps.greenio import pg

    pg.make_asynchronous()

The :func:`make_asynchronous` set the :func:`psycopg2_wait_callback` function
as the waiting callback in psycopg2.

.. seealso::

    Check out how the :mod:`~pulsar.apps.pulse` application uses greenlet
    support to asynchrnously execute :django:`django <>` middleware when using
    :postgresql:`PostgreSql <>` as backend database.


.. autofunction:: psycopg2_wait_callback

.. autofunction:: make_asynchronous

"""
from psycopg2 import ProgrammingError, OperationalError
from psycopg2 import extensions

from pulsar import ImproperlyConfigured

from . import wait_fd


def psycopg2_wait_callback(conn):
    """A wait callback to allow greenlet to work with Psycopg.
    The caller must be from a greenlet other than the main one.
    """
    while 1:
        state = conn.poll()
        if state == extensions.POLL_OK:
            # Done with waiting
            break
        elif state == extensions.POLL_READ:
            print('reading')
            wait_fd(conn)
        elif state == extensions.POLL_WRITE:
            print('writing')
            wait_fd(conn, read=False)
        else:
            raise OperationalError("Bad result from poll: %r" % state)


def make_asynchronous():
    try:
        extensions.POLL_OK
    except AttributeError:
        raise ImproperlyConfigured(
            'Psycopg2 does not have support for asynchronous connections. '
            'You need at least version 2.2.0 of Psycopg2.')
    extensions.set_wait_callback(psycopg2_wait_callback)



'''
The :mod:`pulsar.apps.wsgi.utils` module include several utilities used
by various components in the :ref:`wsgi application <apps-wsgi>`
'''
import time
import re
import textwrap
import logging
from datetime import datetime, timedelta
from email.utils import formatdate


from pulsar import format_traceback
from pulsar.utils.system import json
from pulsar.utils.structures import MultiValueDict
from pulsar.utils.html import escape
from pulsar.utils.pep import to_string
from pulsar.utils.httpurl import (has_empty_content, REDIRECT_CODES, iteritems,
                                  parse_qsl, HTTPError, parse_dict_header,
                                  JSON_CONTENT_TYPES)

from .structures import Accept, RequestCacheControl
from .content import Html, HtmlDocument

__all__ = ['handle_wsgi_error',
           'render_error_debug',
           'wsgi_request',
           'set_wsgi_request_class',
           'dump_environ',
           'HOP_HEADERS']

DEFAULT_RESPONSE_CONTENT_TYPES = ('text/html', 'text/plain'
                                  ) + JSON_CONTENT_TYPES
HOP_HEADERS = frozenset(('connection',
                         'keep-alive',
                         'proxy-authenticate',
                         'proxy-authorization',
                         'te',
                         'trailers',
                         'transfer-encoding',
                         'upgrade')
                        )

logger = logging.getLogger('pulsar.wsgi')
error_css = '''
.pulsar-error {
    width: 500px;
    margin: 50px auto;
}
'''

_RequestClass = None


def wsgi_request(environ, app_handler=None, urlargs=None):
    global _RequestClass
    return _RequestClass(environ, app_handler=app_handler, urlargs=urlargs)


def set_wsgi_request_class(RequestClass):
    global _RequestClass
    _RequestClass = RequestClass


def cookie_date(epoch_seconds=None):
    """Formats the time to ensure compatibility with Netscape's cookie
    standard.

    Accepts a floating point number expressed in seconds since the epoch in, a
    datetime object or a timetuple.  All times in UTC.  The :func:`parse_date`
    function can be used to parse such a date.

    Outputs a string in the format ``Wdy, DD-Mon-YYYY HH:MM:SS GMT``.

    :param expires: If provided that date is used, otherwise the current.
    """
    rfcdate = formatdate(epoch_seconds)
    return '%s-%s-%s GMT' % (rfcdate[:7], rfcdate[8:11], rfcdate[12:25])


def set_cookie(cookies, key, value='', max_age=None, expires=None, path='/',
               domain=None, secure=False, httponly=False):
    '''Set a cookie key into the cookies dictionary *cookies*.'''
    cookies[key] = value
    if expires is not None:
        if isinstance(expires, datetime):
            delta = expires - expires.utcnow()
            # Add one second so the date matches exactly (a fraction of
            # time gets lost between converting to a timedelta and
            # then the date string).
            delta = delta + timedelta(seconds=1)
            # Just set max_age - the max_age logic will set expires.
            expires = None
            max_age = max(0, delta.days * 86400 + delta.seconds)
        else:
            cookies[key]['expires'] = expires
    if max_age is not None:
        cookies[key]['max-age'] = max_age
        # IE requires expires, so set it if hasn't been already.
        if not expires:
            cookies[key]['expires'] = cookie_date(time.time() + max_age)
    if path is not None:
        cookies[key]['path'] = path
    if domain is not None:
        cookies[key]['domain'] = domain
    if secure:
        cookies[key]['secure'] = True
    if httponly:
        cookies[key]['httponly'] = True


_accept_re = re.compile(r'([^\s;,]+)(?:[^,]*?;\s*q=(\d*(?:\.\d+)?))?')


def parse_accept_header(value, cls=None):
    """Parses an HTTP Accept-* header.  This does not implement a complete
    valid algorithm but one that supports at least value and quality
    extraction.

    Returns a new :class:`Accept` object (basically a list of
    ``(value, quality)`` tuples sorted by the quality with some additional
    accessor methods).

    The second parameter can be a subclass of :class:`Accept` that is created
    with the parsed values and returned.

    :param value: the accept header string to be parsed.
    :param cls: the wrapper class for the return value (can be
                         :class:`Accept` or a subclass thereof)
    :return: an instance of `cls`.
    """
    if cls is None:
        cls = Accept
    if not value:
        return cls(None)
    result = []
    for match in _accept_re.finditer(value):
        quality = match.group(2)
        if not quality:
            quality = 1
        else:
            quality = max(min(float(quality), 1), 0)
        result.append((match.group(1), quality))
    return cls(result)


def parse_cache_control_header(value, on_update=None, cls=None):
    """Parse a cache control header.  The RFC differs between response and
    request cache control, this method does not.  It's your responsibility
    to not use the wrong control statements.

    :param value: a cache control header to be parsed.
    :param on_update: an optional callable that is called every time a value
                      on the :class:`~werkzeug.datastructures.CacheControl`
                      object is changed.
    :param cls: the class for the returned object.  By default
                :class:`pulsar.apps.wsgi.structures.RequestCacheControl` is
                used.
    :return: a `cls` object.
    """
    if cls is None:
        cls = RequestCacheControl
    if not value:
        return cls(None, on_update)
    return cls(parse_dict_header(value), on_update)


def _gen_query(query_string, encoding):
    # keep_blank_values=True
    for key, value in parse_qsl((query_string or ''), True):
        yield (to_string(key, encoding, errors='replace'),
               to_string(value, encoding, errors='replace'))


def query_dict(query_string, encoding='utf-8'):
    if query_string:
        return dict(MultiValueDict(_gen_query(query_string, encoding)).items())
    else:
        return {}


error_messages = {
    500: 'An exception has occurred while evaluating your request.',
    404: 'Cannot find what you are looking for.'
}


class dump_environ(object):
    __slots__ = ('environ',)

    def __init__(self, environ):
        self.environ = environ

    def __str__(self):
        def _():
            for k, v in iteritems(self.environ):
                try:
                    v = str(v)
                except Exception as e:
                    v = str(e)
                yield '%s=%s' % (k, v)
        return '\n%s\n' % '\n'.join(_())


def handle_wsgi_error(environ, exc):
    '''The default error handler while serving a WSGI request.

    :param environ: The WSGI environment.
    :param exc: the exception
    :return: a :class:`.WsgiResponse`
    '''
    if isinstance(exc, tuple):
        exc_info = exc
        exc = exc[1]
    else:
        exc_info = True
    request = wsgi_request(environ)
    request.cache.handle_wsgi_error = True
    response = request.response
    if isinstance(exc, HTTPError):
        response.status_code = exc.code or 500
    else:
        response.status_code = getattr(exc, 'status', 500)
        response.headers.update(getattr(exc, 'headers', None) or ())
    path = '@ %s "%s"' % (request.method, request.path)
    status = response.status_code
    if status == 500:
        logger.critical('Unhandled exception during HTTP response %s.%s',
                        path, dump_environ(environ), exc_info=exc_info)
    else:
        msg = str(exc)
        msg = '' if not msg else ' - %s' % msg
        logger.warning('HTTP %s %s%s', response.status, path, msg)
    if has_empty_content(status, request.method) or status in REDIRECT_CODES:
        response.content_type = None
        response.content = None
    else:
        request.cache.pop('html_document', None)
        renderer = environ.get('error.handlers', {}).get(status, render_error)
        try:
            content = renderer(request, exc)
        except Exception:
            logger.critical('Error while rendering error', exc_info=True)
            response.content_type = 'text/plain'
            content = 'Critical server error'
        if content is not response:
            response.content = content
    return response


def render_error(request, exc):
    '''Default renderer for errors.'''
    cfg = request.get('pulsar.cfg')
    debug = cfg.debug if cfg else False
    response = request.response
    if not response.content_type:
        response.content_type = request.content_types.best_match(
            DEFAULT_RESPONSE_CONTENT_TYPES)
    content_type = None
    if response.content_type:
        content_type = response.content_type.split(';')[0]
    is_html = content_type == 'text/html'

    if debug:
        msg = render_error_debug(request, exc, is_html)
    else:
        msg = error_messages.get(response.status_code) or str(exc)
        if is_html:
            msg = textwrap.dedent("""
                <h1>{0[reason]}</h1>
                {0[msg]}
                <h3>{0[version]}</h3>
            """).format({"reason": response.status, "msg": msg,
                         "version": request.environ['SERVER_SOFTWARE']})
    #
    if content_type == 'text/html':
        doc = HtmlDocument(title=response.status)
        doc.head.links.append('bootstrap_css')
        doc.head.embedded_css.append(error_css)
        doc.body.append(Html('div', msg, cn='pulsar-error'))
        return doc.render(request)
    elif content_type in JSON_CONTENT_TYPES:
        return json.dumps({'status': response.status_code,
                           'message': msg})
    else:
        return '\n'.join(msg) if isinstance(msg, (list, tuple)) else msg


def render_error_debug(request, exception, is_html):
    '''Render the ``exception`` traceback
    '''
    error = Html('div', cn='well well-lg') if is_html else []
    for trace in format_traceback(exception):
        counter = 0
        for line in trace.split('\n'):
            if line.startswith('  '):
                counter += 1
                line = line[2:]
            if line:
                if is_html:
                    line = Html('p', escape(line), cn='text-danger')
                    if counter:
                        line.css({'margin-left': '%spx' % (20*counter)})
                error.append(line)
    if is_html:
        error = Html('div', Html('h1', request.response.status), error)
    return error



# Modified version of skiplist
# http://code.activestate.com/recipes/
#    576930-efficient-running-median-using-an-indexable-skipli/
import sys
from random import random
from math import log
from collections import Sequence

ispy3k = int(sys.version[0]) >= 3

if not ispy3k:
    range = xrange


class Node(object):
    __slots__ = ('score', 'value', 'next', 'width')

    def __init__(self, score, value, next, width):
        self.score, self.value, self.next, self.width = (score, value,
                                                         next, width)


SKIPLIST_MAXLEVEL = 32     # Should be enough for 2^32 elements

neg_inf = float('-inf')
inf = float('inf')


class Skiplist(Sequence):
    '''Sorted collection supporting O(log n) insertion,
    removal, and lookup by rank.'''
    __slots__ = ('_unique', '_size', '_head', '_level')

    def __init__(self, data=None, unique=False):
        self._unique = unique
        self.clear()
        if data is not None:
            self.extend(data)

    def __repr__(self):
        return list(self).__repr__()

    def __str__(self):
        return self.__repr__()

    def __len__(self):
        return self._size

    def __getitem__(self, index):
        node = self._head
        traversed = 0
        index += 1
        for i in range(self._level-1, -1, -1):
            while node.next[i] and (traversed + node.width[i]) <= index:
                traversed += node.width[i]
                node = node.next[i]
            if traversed == index:
                return node.value
        raise IndexError('skiplist index out of range')

    def clear(self):
        '''Clear the container from all data.'''
        self._size = 0
        self._level = 1
        self._head = Node('HEAD', None,
                          [None]*SKIPLIST_MAXLEVEL,
                          [1]*SKIPLIST_MAXLEVEL)

    def extend(self, iterable):
        '''Extend this skiplist with an iterable over
        ``score``, ``value`` pairs.
        '''
        i = self.insert
        for score_values in iterable:
            i(*score_values)
    update = extend

    def rank(self, score):
        '''Return the 0-based index (rank) of ``score``.

        If the score is not available it returns a negative integer which
        absolute score is the right most closest index with score less than
        ``score``.
        '''
        node = self._head
        rank = 0
        for i in range(self._level-1, -1, -1):
            while node.next[i] and node.next[i].score < score:
                rank += node.width[i]
                node = node.next[i]
        node = node.next[0]
        if node and node.score == score:
            return rank
        else:
            return -2 - rank

    def range(self, start=0, end=None, scores=False):
        N = len(self)
        if start < 0:
            start = max(N + start, 0)
        if start >= N:
            raise StopIteration
        if end is None:
            end = N
        elif end < 0:
            end = max(N + end, 0)
        else:
            end = min(end, N)
        if start >= end:
            raise StopIteration
        node = self._head.next[0]
        index = 0
        while node:
            if index >= start:
                if index < end:
                    yield (node.score, node.value) if scores else node.value
                else:
                    break
            index += 1
            node = node.next[0]

    def range_by_score(self, minval, maxval, include_min=True,
                       include_max=True, start=0, num=None, scores=False):
        node = self._head
        if include_min:
            for i in range(self._level-1, -1, -1):
                while node.next[i] and node.next[i].score < minval:
                    node = node.next[i]
        else:
            for i in range(self._level-1, -1, -1):
                while node.next[i] and node.next[i].score <= minval:
                    node = node.next[i]
        node = node.next[0]
        if node and node.score >= minval:
            index = 0
            while node:
                index += 1
                if num is not None and index - start > num:
                    break
                if ((include_max and node.score > maxval) or
                        (not include_max and node.score >= maxval)):
                    break
                if index > start:
                    yield (node.score, node.value) if scores else node.value
                node = node.next[0]

    def insert(self, score, value):
        # find first node on each level where node.next[levels].score > score
        if score != score:
            raise ValueError('Cannot insert score {0}'.format(score))
        chain = [None] * SKIPLIST_MAXLEVEL
        rank = [0] * SKIPLIST_MAXLEVEL
        node = self._head
        for i in range(self._level-1, -1, -1):
            # store rank that is crossed to reach the insert position
            rank[i] = 0 if i == self._level-1 else rank[i+1]
            while node.next[i] and node.next[i].score <= score:
                rank[i] += node.width[i]
                node = node.next[i]
            chain[i] = node
        # the score already exist
        if chain[0].score == score and self._unique:
            return
        # insert a link to the newnode at each level
        level = min(SKIPLIST_MAXLEVEL, 1 - int(log(random(), 2.0)))
        if level > self._level:
            for i in range(self._level, level):
                rank[i] = 0
                chain[i] = self._head
                chain[i].width[i] = self._size
            self._level = level

        # create the new node
        node = Node(score, value, [None]*level, [None]*level)
        for i in range(level):
            prevnode = chain[i]
            steps = rank[0] - rank[i]
            node.next[i] = prevnode.next[i]
            node.width[i] = prevnode.width[i] - steps
            prevnode.next[i] = node
            prevnode.width[i] = steps + 1

        # increment width for untouched levels
        for i in range(level, self._level):
            chain[i].width[i] += 1

        self._size += 1
        return node

    def remove_range(self, start, end, callback=None):
        '''Remove a range by rank.

        This is equivalent to perform::

            del l[start:end]

        on a python list.
        It returns the number of element removed.
        '''
        N = len(self)
        if start < 0:
            start = max(N + start, 0)
        if start >= N:
            return 0
        if end is None:
            end = N
        elif end < 0:
            end = max(N + end, 0)
        else:
            end = min(end, N)
        if start >= end:
            return 0
        node = self._head
        index = 0
        chain = [None] * self._level
        for i in range(self._level-1, -1, -1):
            while node.next[i] and (index + node.width[i]) <= start:
                index += node.width[i]
                node = node.next[i]
            chain[i] = node
        node = node.next[0]
        initial = self._size
        while node and index < end:
            next = node.next[0]
            self._remove_node(node, chain)
            index += 1
            if callback:
                callback(node.score, node.value)
            node = next
        return initial - self._size

    def remove_range_by_score(self, minval, maxval, include_min=True,
                              include_max=True, callback=None):
        '''Remove a range with scores between ``minval`` and ``maxval``.

        :param minval: the start value of the range to remove
        :param maxval: the end value of the range to remove
        :param include_min: whether or not to include ``minval`` in the
            values to remove
        :param include_max: whether or not to include ``maxval`` in the
            scores to to remove
        :param callback: optional callback function invoked for each
            score, value pair removed.
        :return: the number of elements removed.
        '''
        node = self._head
        chain = [None] * self._level
        if include_min:
            for i in range(self._level-1, -1, -1):
                while node.next[i] and node.next[i].score < minval:
                    node = node.next[i]
                chain[i] = node
        else:
            for i in range(self._level-1, -1, -1):
                while node.next[i] and node.next[i].score <= minval:
                    node = node.next[i]
                chain[i] = node
        node = node.next[0]
        initial = self._size
        while node and node.score >= minval:
            if ((include_max and node.score > maxval) or
                    (not include_max and node.score >= maxval)):
                break
            next = node.next[0]
            self._remove_node(node, chain)
            if callback:
                callback(node.score, node.value)
            node = next
        return initial - self._size

    def count(self, minval, maxval, include_min=True, include_max=True):
        '''Returns the number of elements in the skiplist with a score
        between min and max.
        '''
        rank1 = self.rank(minval)
        if rank1 < 0:
            rank1 = -rank1 - 1
        elif not include_min:
            rank1 += 1
        rank2 = self.rank(maxval)
        if rank2 < 0:
            rank2 = -rank2 - 1
        elif include_max:
            rank2 += 1
        return max(rank2 - rank1, 0)

    def __iter__(self):
        'Iterate over values in sorted order'
        node = self._head.next[0]
        while node:
            yield node.score, node.value
            node = node.next[0]

    def flat(self):
        return tuple(self._flat())

    def _flat(self):
        node = self._head.next[0]
        while node:
            yield node.score
            yield node.value
            node = node.next[0]

    def _remove_node(self, node, chain):
        for i in range(self._level):
            if chain[i].next[i] == node:
                chain[i].width[i] += node.width[i] - 1
                chain[i].next[i] = node.next[i]
            else:
                chain[i].width[i] -= 1
        self._size -= 1


class SkipListSlice(object):
    __slots__ = ('sl', 'slice')

    def __init__(self, sl, slice):
        self.sl = sl
        self.slice = slice



'''Sphinx extension for displaying Setting information
'''
from inspect import isfunction

from sphinx.util.compat import Directive
from docutils import nodes, statemachine

from pulsar.utils.config import ordered_settings, section_docs
import pulsar.apps.wsgi
import pulsar.apps.tasks
import pulsar.apps.test

targetid = "pulsar_settings"


class pulsar_settings(nodes.General, nodes.Element):
    pass


class PulsarSettings(Directive):
    has_content = False
    required_arguments = 0

    def sections(self):
        sec = {}
        sections = []
        for sett in ordered_settings():
            s = sett.section
            if s not in sec:
                sections.append(s)
                sec[s] = [sett]
            else:
                sec[s].append(sett)
        for s in sections:
            yield s, sec[s]

    def text(self):
        for section, settings in self.sections():
            if section is None:
                continue
            s = section.lower().replace(' ', '-')
            yield '.. _setting-section-{0}:\n\n\
{1}\n=====================================\n\n'.format(s, section)
            section_doc = section_docs.get(section)
            if section_doc:
                yield section_doc
                yield '\n'
            for sett in settings:
                desc = sett.desc.strip()
                yield '.. _setting-{0}:\n\n\
{0}\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n\n'.format(sett.name)
                if sett.app:
                    yield '*Application namespace*: ``%s``\n' % sett.app
                yield '*Config name*: ``%s``\n' % sett.name
                if sett.flags:
                    yield '*Command line*: %s\n' %\
                                ', '.join('``%s``' % f for f in sett.flags)
                if isfunction(sett.default):
                    default = ':func:`%s`' % sett.default.__name__
                else:
                    default = '``%r``' % sett.default
                yield '*Default*: %s\n' % default
                yield desc + '\n'

    def run(self):
        env = self.state.document.settings.env
        rawdocs = '\n'.join(self.text())

        source = self.state_machine.input_lines.source(
            self.lineno - self.state_machine.input_offset - 1)

        encoding = self.options.get(
            'encoding', self.state.document.settings.input_encoding)
        tab_width = self.options.get(
            'tab-width', self.state.document.settings.tab_width)


        if 'literal' in self.options:
            # Convert tabs to spaces, if `tab_width` is positive.
            if tab_width >= 0:
                text = rawtext.expandtabs(tab_width)
            else:
                text = rawtext
            literal_block = nodes.literal_block(rawtext, text, source=path)
            literal_block.line = 1
            return [literal_block]
        else:
            include_lines = statemachine.string2lines(
                rawdocs, tab_width, convert_whitespace=1)
            self.state_machine.insert_input(include_lines, targetid)
            return []


def setup(app):
    app.add_directive(targetid, PulsarSettings)
